Package: schmear
Title: Build Structured Data Frame Subtypes
Version: 0.1.0
Authors@R: 
    person("Cory", "McCartan", , "mccartan@psu.edu", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0002-6251-669X"))
Description: Provides developer-focused helper functions and S3 classes to ease
    the creation of structured subtypes of data frames. Developers can require
    certain columns and types to be present, and can enforce crossing and 
    nesting relationships between values in different columns. Type-specific
    metadata and attributes are preserved through common data frame
    manipulations.
Imports: rlang, cli, vctrs
Suggests: dplyr, pillar, testthat (>= 3.0.0)
License: MIT + file LICENSE
Encoding: UTF-8
URL: https://corymccartan.com/schmear/,
        https://github.com/CoryMcCartan/schmear
BugReports: https://github.com/CoryMcCartan/schmear/issues
Config/testthat/edition: 3
Config/roxygen/version: 8.0.0
Config/roxygen/markdown: TRUE
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-22 23:34:42 UTC; cmccartan
Author: Cory McCartan [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-6251-669X>)
Maintainer: Cory McCartan <mccartan@psu.edu>
Repository: CRAN
Date/Publication: 2026-05-29 09:30:11 UTC
Built: R 4.5.2; ; 2026-05-29 10:19:14 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("[",sch_df)
S3method("names<-",sch_df)
S3method(dplyr::dplyr_col_modify,sch_df)
S3method(dplyr::dplyr_reconstruct,sch_df)
S3method(dplyr::dplyr_row_slice,sch_df)
S3method(format,sch_schema)
S3method(format,sch_type)
S3method(pillar::tbl_sum,sch_df)
S3method(print,sch_schema)
S3method(print,sch_type)
export(new_sch_df)
export(sch_any)
export(sch_character)
export(sch_coerce)
export(sch_custom)
export(sch_date)
export(sch_datetime)
export(sch_factor)
export(sch_inherits)
export(sch_integer)
export(sch_list_of)
export(sch_logical)
export(sch_multiple)
export(sch_nest)
export(sch_numeric)
export(sch_others)
export(sch_schema)
export(sch_validate)
export(validate_sch_df)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# schmear <a href="https://corymccartan.com/schmear/"><img src="man/figures/logo.png" align="right" height="138" alt="schmear logo" /></a>

<!-- badges: start -->

[![Codecov test
coverage](https://codecov.io/gh/CoryMcCartan/schmear/graph/badge.svg)](https://app.codecov.io/gh/CoryMcCartan/schmear)
[![R-CMD-check](https://github.com/CoryMcCartan/schmear/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/CoryMcCartan/schmear/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

Provides developer-focused helper functions and S3 classes to ease the
creation of structured subtypes of data frames which require certain
columns and types to be present, and which preserve special attributes
through common data frame manipulations.

## Installation

You can install the released version of schmear from CRAN with:

``` r
install.packages("schmear")
```

You can install the development version from
[GitHub](https://github.com/CoryMcCartan/schmear) with:

``` r
# install.packages("pak")
pak::pak("CoryMcCartan/schmear")
```

# Example: Election data schema
#
# Relationship structure:
#   (state + election + jurisdiction + contest + district) /
#       (geo * (party / candidate)) /
#       (time * method)
#
# The race ID is a compound key of (state, election, jurisdiction, contest,
# district). Within each race, geo x party x time x method are fully crossed.
# Candidate is nested within party. Stage and votes are outcome columns.

library(schmear)

# Constants --------------------------------------------------------------------

STATES <- c(state.abb, "DC")

CONTESTS <- c(
    "pres", "sen", "house", "gov", "ag", "ltgov", "stsen", "sthouse",
    "mayor", "prop"
)

PARTIES <- c("dem", "rep", "ind", "oth", "all", "yes", "no")

VOTE_METHODS <- c("abs", "adv", "mail", "early", "eday", "prov", "all")

# Schema -----------------------------------------------------------------------

# stage can be integer (editorial stage code) or character/factor stage label
sch_stage <- sch_custom(
    name = "stage",
    desc = "Reporting stage",
    check = function(x, type) is.integer(x) || is.character(x) || is.factor(x),
    msg = function(type) "integer stage code or character/factor stage label",
    coerce = function(x, type) x
)

schema_elec <- sch_schema(
    .desc = "Tidy election data",
    .relationships = ~ (state + election + jurisdiction + contest + district) /
        (geo * (party / candidate)) /
        (time * method),
    # --- Race identifier columns ---
    state = sch_factor("State USPS abbreviation", levels = STATES, strict = FALSE, missing = FALSE),
    election = sch_date("Election date", missing = FALSE),
    jurisdiction = sch_factor("Jurisdiction identifier", strict = FALSE, missing = FALSE),
    contest = sch_factor("Contest type", levels = CONTESTS, strict = FALSE, missing = FALSE),
    district = sch_integer("District number", bounds = c(1, Inf)),
    # --- Within-race columns ---
    geo = sch_factor("Geographic unit", strict = FALSE, missing = FALSE),
    party = sch_factor("Party abbreviation", levels = PARTIES, strict = FALSE, missing = FALSE),
    candidate = sch_factor("Candidate name", strict = FALSE),
    time = sch_datetime("Data record timestamp", missing = FALSE),
    method = sch_factor("Vote method", levels = VOTE_METHODS, strict = FALSE, missing = FALSE),
    # --- Outcome columns ---
    stage = sch_stage,
    votes = sch_numeric("Vote count", bounds = c(0, Inf), missing = FALSE),
    past = sch_numeric(
        "Past election vote baseline",
        bounds = c(0, Inf),
        missing = FALSE,
        required = FALSE
    ),
    sch_others()
)

print(schema_elec)

# Compliant data ---------------------------------------------------------------
#
# Structure:
#   2 races: NJ pres (dist 1), PA pres (dist 1)
#   2 geos per race
#   2 candidates (nested within party)
#   2 time snapshots x 2 methods per (race, geo, candidate)
#   Total: 2 races x 2 geos x 2 candidates x 2 times x 2 methods = 32 rows

elec_date <- as.Date("2024-11-05")
t1 <- as.POSIXct("2024-11-05 23:00:00", tz = "America/New_York")
t2 <- as.POSIXct("2024-11-05 23:30:00", tz = "America/New_York")

make_race_geo <- function(state, geo, base_votes) {
    expand.grid(
        candidate = c("Harris, Kamala D.", "Trump, Donald J."),
        time = c(t1, t2),
        method = c("eday", "all"),
        stringsAsFactors = FALSE
    ) |>
        transform(
            state = state,
            election = elec_date,
            jurisdiction = "US",
            contest = "pres",
            district = 1L,
            geo = geo,
            party = ifelse(candidate == "Harris, Kamala D.", "dem", "rep"),
            stage = ifelse(time == t1, 4L, 10L),
            votes = base_votes + seq_len(8) * 100
        )
}

df <- rbind(
    make_race_geo("NJ", "34001", 12000),
    make_race_geo("NJ", "34003", 22000),
    make_race_geo("PA", "42001", 8000),
    make_race_geo("PA", "42003", 55000)
)
rownames(df) <- NULL

cat("\nCompliant data (", nrow(df), "rows):\n")
print(head(df, 12))
cat("\nValidating compliant data...\n")
sch_validate(schema_elec, df)
cat("OK\n")

# Corruptions ------------------------------------------------------------------

cat("\n--- Corruption 1: invalid party value ---\n")
bad1 <- df
bad1$party[1] <- "lib"
try(sch_validate(schema_elec, bad1))

cat("\n--- Corruption 2: negative vote count ---\n")
bad2 <- df
bad2$votes[1] <- -500
try(sch_validate(schema_elec, bad2))

cat("\n--- Corruption 3: missing required 'votes' column ---\n")
bad3 <- df[, setdiff(names(df), "votes")]
try(sch_validate(schema_elec, bad3))

cat("\n--- Corruption 4: NA in votes (not permitted) ---\n")
bad4 <- df
bad4$votes[1] <- NA_real_
try(sch_validate(schema_elec, bad4))

cat("\n--- Corruption 5: duplicate primary key (duplicate row) ---\n")
bad5 <- rbind(df, df[1, ])
try(sch_validate(schema_elec, bad5))

cat("\n--- Corruption 6: incomplete crossing (missing candidate in one geo) ---\n")
bad6 <- df[!(df$geo == "34003" & df$party == "rep"), ]
try(sch_validate(schema_elec, bad6))

cat("\n--- Corruption 7: incomplete crossing (one geo has extra time snapshot) ---\n")
extra <- make_race_geo("NJ", "34001", 18000)
t3 <- as.POSIXct("2024-11-06 02:00:00", tz = "America/New_York")
extra$time <- t3
bad7 <- rbind(df, extra)
try(sch_validate(schema_elec, bad7))

# Example: MCMC draws schema
#
# Structure: chain x draw x param fully crossed.
# Every (chain, draw, param) combination exists exactly once.

library(schmear)

# Schema -----------------------------------------------------------------------

schema <- sch_schema(
    .desc = "MCMC posterior draws",
    .relationships = ~ chain * draw * param,
    chain = sch_integer("Chain number", bounds = c(1, Inf)),
    draw = sch_integer("Draw number within chain", bounds = c(1, Inf)),
    param = sch_factor(
        "Parameter name",
        levels = c("mu", paste0("beta", 1:20), "sigma", "log_lik"),
        strict = FALSE
    ),
    value = sch_numeric("Parameter draw value")
)

print(schema)

# Compliant data ---------------------------------------------------------------

n_chains <- 4L
n_draws <- 10000L
params <- c("mu", paste0("beta", 1:20), "sigma", "log_lik")

draws_df <- expand.grid(
    chain = seq_len(n_chains),
    draw = seq_len(n_draws),
    param = params,
    stringsAsFactors = FALSE
)
draws_df$chain <- as.integer(draws_df$chain)
draws_df$draw <- as.integer(draws_df$draw)
draws_df$param <- factor(draws_df$param, levels = params)
draws_df$value <- c(
    rnorm(n_chains * n_draws),                  # mu draws
    rnorm(n_chains * n_draws * 20, 0, 0.5),     # beta draws
    abs(rnorm(n_chains * n_draws, 1, 0.2)),     # sigma draws (positive)
    rnorm(n_chains * n_draws, -200, 10)         # log_lik draws
)

draws_df <- draws_df[order(draws_df$chain, draws_df$draw, draws_df$param), ]
rownames(draws_df) <- NULL

cat("\nCompliant data (first 12 rows):\n")
print(head(draws_df, 12))
cat("\nValidating compliant data...\n")
sch_validate(schema, draws_df)
cat("OK\n")

# Corruptions ------------------------------------------------------------------

cat("\n--- Corruption 1: wrong type for 'value' (character instead of numeric) ---\n")
bad1 <- draws_df
bad1$value <- as.character(bad1$value)
try(sch_validate(schema, bad1))

cat("\n--- Corruption 2: missing required 'chain' column ---\n")
bad2 <- draws_df[, c("draw", "param", "value")]
try(sch_validate(schema, bad2))

cat("\n--- Corruption 3: duplicate (chain, draw, param) combo ---\n")
bad3 <- draws_df
bad3$draw[bad3$chain == 1L & bad3$draw == 100L] <- 1L
try(sch_validate(schema, bad3))

cat("\n--- Corruption 4: incomplete crossing (chain 2 missing log_lik) ---\n")
bad4 <- draws_df[!(draws_df$chain == 2L & draws_df$param == "log_lik"), ]
try(sch_validate(schema, bad4))

cat("\n--- Corruption 5: out-of-bounds draw number (draw = 0) ---\n")
bad5 <- draws_df
bad5$draw[bad5$chain == 1L & bad5$draw == 1L] <- 0L
try(sch_validate(schema, bad5))

cat("\n--- Corruption 6: NA in value column ---\n")
bad6 <- draws_df
bad6$value[1] <- NA_real_
# value allows missing by default, so this should PASS
sch_validate(schema, bad6)
cat("(Passed - 'value' allows NAs by default)\n")

# Stricter schema that forbids NAs in value
schema_strict <- sch_schema(
    .desc = "MCMC posterior draws (strict: no NA values)",
    .relationships = ~ chain * draw * param,
    chain = sch_integer("Chain number", bounds = c(1, Inf)),
    draw = sch_integer("Draw number within chain", bounds = c(1, Inf)),
    param = sch_factor(
        "Parameter name",
        levels = c("mu", "sigma", "log_lik"),
        strict = FALSE
    ),
    value = sch_numeric("Parameter draw value", missing = FALSE)
)

try(sch_validate(schema_strict, bad6))

# Example: Ecological inference specification (matching 'seine' package)
#
# An ei_spec data frame has one row per geographic unit (e.g., county).
# It contains:
#   - predictor columns (racial composition shares, summing to 1 per row)
#   - outcome columns (vote share estimates, summing to 1 per row)
#   - a `total` column (number of individuals per unit, e.g. voters)
#   - optional covariate columns
#
# The sch_groups attribute maps group names ("predictors", "outcomes") to
# the actual column names in the data frame. Cross-column checks enforce that
# both groups sum to 1 in every row.

library(schmear)

# Schema -----------------------------------------------------------------------

schema <- sch_schema(
    .desc = "Ecological inference specification",
    sch_multiple(
        name = "predictors",
        desc = "Racial composition shares",
        type = sch_numeric(bounds = c(0, 1), missing = FALSE),
        check = function(x, type) {
            row_sums <- Reduce("+", x)
            all(abs(row_sums - 1) < 1e-6)
        },
        msg = function(type) "rows sum to 1",
        coerce = function(x, type) x
    ),
    sch_multiple(
        name = "outcomes",
        desc = "Vote share estimates",
        type = sch_numeric(missing = FALSE),
    ),
    sch_multiple(
        name = "covariates",
        type = sch_numeric(missing = FALSE),
        required = FALSE,
    ),
    total = sch_numeric("Total individuals per unit", bounds = c(1, Inf), missing = FALSE),
    sch_others()
)

print(schema)

# Compliant data ---------------------------------------------------------------

n_units <- 50L

# Generate Dirichlet-like composition data (rows sum exactly to 1)
dirichlet_sample <- function(n, k, alpha = rep(1, k)) {
    raw <- matrix(rgamma(n * k, shape = alpha), nrow = n, byrow = TRUE)
    raw / rowSums(raw)
}

pred_shares <- dirichlet_sample(n_units, 3)
out_shares <- dirichlet_sample(n_units, 3)

df <- data.frame(
    vap_white = pred_shares[, 1],
    vap_black = pred_shares[, 2],
    vap_other = pred_shares[, 3],
    pres_dem = out_shares[, 1],
    pres_rep = out_shares[, 2],
    pres_other = out_shares[, 3],
    total = as.integer(runif(n_units, 500, 50000)),
    income_med = runif(n_units, 30000, 120000)
)
class(df) <- c("ei_spec", "sch_df", "tbl_df", "tbl", "data.frame")

attr(df, "sch_groups") <- list(
    predictors = c("vap_white", "vap_black", "vap_other"),
    outcomes = c("pres_dem", "pres_rep", "pres_other"),
    covariates = "income_med"
)

cat("\nCompliant data (first 5 rows):\n")
print(head(df, 5))
cat("\nValidating compliant data...\n")
sch_validate(schema, df)
cat("OK\n")

# Corruptions ------------------------------------------------------------------

cat("\n--- Corruption 1: missing sch_groups attribute ---\n")
bad1 <- df
attr(bad1, "sch_groups") <- NULL
try(sch_validate(schema, bad1))

cat("\n--- Corruption 2: predictor column out of [0, 1] bounds ---\n")
bad2 <- df
bad2$vap_white[1] <- 1.5
try(sch_validate(schema, bad2))

cat("\n--- Corruption 3: predictor shares do not sum to 1 ---\n")
bad3 <- df
bad3$vap_white[1] <- 0.1
try(sch_validate(schema, bad3))

cat("\n--- Corruption 4: missing 'total' column ---\n")
bad4 <- df[, setdiff(names(df), "total")]
attr(bad4, "sch_groups") <- attr(df, "sch_groups")
try(sch_validate(schema, bad4))

cat("\n--- Corruption 5: total contains zero (below lower bound of 1) ---\n")
bad5 <- df
bad5$total[1] <- 0L
try(sch_validate(schema, bad5))

cat("\n--- Corruption 6: predictor column contains NA ---\n")
bad6 <- df
bad6$vap_white[1] <- NA_real_
try(sch_validate(schema, bad6))

cat("\n--- Corruption 7: sch_groups references a column not present in data ---\n")
bad7 <- df[, setdiff(names(df), "vap_white")]
attr(bad7, "sch_groups") <- attr(df, "sch_groups")
try(sch_validate(schema, bad7))

cat("\n--- Corruption 8: outcomes group absent from sch_groups ---\n")
bad8 <- df
attr(bad8, "sch_groups") <- list(predictors = c("vap_white", "vap_black", "vap_other"))
try(sch_validate(schema, bad8))

cat("\n--- Corruption 9: outcomes are character, not numeric ---\n")
bad9 <- df
bad9$pres_dem <- as.character(bad9$pres_dem)
try(sch_validate(schema, bad9))

#' Construct and validate a schema-aware data frame
#'
#' Bare-bones constructor for a data frame with an attached schema. This function
#' should be called by package developers writing their own internal constructors.
#' The only checks are for the types of `data` and `schema`.
#' The `validate_sch_df()` is a lightweight wrapper around [sch_validate()] that
#' also returns the input.
#'
#' @param data A data frame.
#' @param schema A `sch_schema` object.
#' @param groups A named list of character vectors of column names, for use
#'   with [sch_multiple()].
#' @param class Additional classes to add to the object, in addition to "sch_df" and tibble/data.frame classes.
#' @param use_tbl If `TRUE`, the returned object will have tibble
#'   classes `tbl_df` and `tbl` in addition to `data.frame`. As a reminder,
#'   `tbl_df` affects the behavior of the object (slicing, row names, etc.),
#'   while `tbl` affects printing only.
#'
#' @returns A data frame with class `c(class, "sch_df", ...)` and attributes
#'   `sch_schema` and `sch_groups`.
#'
#' @examples
#' schema = sch_schema(
#'     .desc = "MCMC draws",
#'     .relationships = ~ chain * draw * parameter,
#'     chain = sch_integer("Chain number"),
#'     draw = sch_integer("Draw number", bounds = c(1, Inf), closed = c(TRUE, FALSE)),
#'     parameter = sch_character("Parameter name"),
#'     value = sch_numeric("Parameter value")
#' )
#' d_raw = data.frame(chain = 1L, draw = 1:4, parameter = "mu", value = rnorm(4))
#' d = new_sch_df(d_raw, schema, class="mcmc_draws")
#' validate_sch_df(d)
#' str(d)
#' @name sch_df
NULL

#' @rdname sch_df
#' @export
new_sch_df <- function(data, schema, groups = NULL, class = NULL, use_tbl = TRUE) {
    assert(is.list(data), "{.arg data} must be a data frame or list")
    assert(inherits(schema, "sch_schema"), "{.arg schema} must be a {.cls sch_schema}")

    if (isTRUE(use_tbl)) {
        class = c(class, "sch_df", "tbl_df", "tbl")
    } else {
        class = c(class, "sch_df")
    }

    vctrs::new_data_frame(
        data,
        sch_schema = schema,
        sch_groups = groups,
        class = class
    )
}

#' @param x A `sch_df` object.
#' @returns The input, if validation is successful.
#'
#' @rdname sch_df
#' @export
validate_sch_df <- function(x) {
    sch_validate(attr(x, "sch_schema"), x)
    x
}


#' @exportS3Method pillar::tbl_sum
tbl_sum.sch_df <- function(x, ...) {
    sch = attr(x, "sch_schema")
    pre = paste(length(sch$cols), "column rules")
    desc = attr(sch, "desc")
    names(pre) = if (is.null(desc)) "<sch_df>" else desc
    c(pre, NextMethod())
}

#' dplyr integration for sch_df
#'
#' These methods hook into dplyr's three extension generics
#' ([dplyr::dplyr_row_slice()], [dplyr::dplyr_col_modify()],
#' [dplyr::dplyr_reconstruct()]) plus the base-R [names()] replacement and
#' 1-d `[` to enforce schema constraints across dplyr operations with near-zero
#' overhead.
#'
#' Each method calls `NextMethod()` first (performing the actual dplyr
#' operation) and then re-validates the result against the schema, running only
#' the checks that can plausibly be violated by that type of operation:
#'
#' | Operation | Checks run |
#' |---|---|
#' | `dplyr_row_slice` | none by default; `"relationships"` behind a flag |
#' | `dplyr_col_modify` | `"names"`, `"types"`, `"distinct"` |
#' | `dplyr_reconstruct` | `"names"`, `"types"` |
#' | `[` (1-d column subsetting) | `"names"` |
#' | `names<-` | errors if any *schema* column name is changed |
#'
#' @section Row slicing (`arrange`, `filter`, `slice`, semi/anti joins):
#' Row operations cannot introduce new name or type violations, so no
#' validation is run by default. Relationship constraints (crossing
#' completeness, primary-key uniqueness) *can* be broken by removing rows.
#' Pass `.check_relationships = TRUE` in `...` to opt in to that check.
#'
#' @section Column modification (`mutate`):
#' A column modification can delete required columns (via `NULL` assignment),
#' assign the wrong type, or introduce duplicate values into a `distinct = TRUE`
#' column, so all three cheap checks are run.
#'
#' @section Reconstruction (joins):
#' `dplyr_reconstruct()` is called after joins. Only names and types are
#' checked: distinct and relationship checks are omitted because joins can
#' intentionally produce non-distinct rows or incomplete crossings.
#'
#' @section Column subsetting (`select`, `relocate`):
#' A 1-d `[` call selects or reorders columns. Reordering is always safe; but
#' selecting a column subset could drop required columns, so a names check is
#' run.
#'
#' @section Renaming (`rename`, `rename_with`, `select` with rename):
#' Renaming a column that belongs to the schema (either directly or as a member
#' of a [sch_multiple()] group) is never valid without also updating the schema
#' definition, so an error is raised immediately.
#'
#' @name dplyr-sch_df
NULL


# dplyr generics ----------------------------------------------------------

#' @exportS3Method dplyr::dplyr_row_slice
dplyr_row_slice.sch_df <- function(data, i, ..., .check_relationships = FALSE) {
    out <- NextMethod()
    if (.check_relationships) {
        sch_validate(
            attr(data, "sch_schema"),
            out,
            check = "relationships",
            call = rlang::caller_env()
        )
    }
    out
}

#' @exportS3Method dplyr::dplyr_col_modify
dplyr_col_modify.sch_df <- function(data, cols) {
    out <- NextMethod()
    sch_validate(
        attr(data, "sch_schema"),
        out,
        check = c("names", "types", "distinct"),
        call = rlang::caller_env()
    )
    out
}

#' @exportS3Method dplyr::dplyr_reconstruct
dplyr_reconstruct.sch_df <- function(data, template) {
    out <- NextMethod()
    sch_validate(
        attr(template, "sch_schema"),
        out,
        check = c("names", "types"),
        call = rlang::caller_env()
    )
    out
}


# Base-R generics ---------------------------------------------------------

#' @export
`[.sch_df` <- function(x, i, j, ..., drop = FALSE) {
    out <- NextMethod()
    # Only validate on 1-d (column) indexing. nargs() == 2 means x[i] with no j.
    if (nargs() == 2L && inherits(out, "sch_df")) {
        sch_validate(
            attr(x, "sch_schema"),
            out,
            check = "names",
            call = rlang::caller_env()
        )
    }
    out
}

#' @export
`names<-.sch_df` <- function(x, value) {
    old_names <- names(x)
    schema <- attr(x, "sch_schema")
    groups <- attr(x, "sch_groups")

    # Columns that map directly to data frame columns (exclude sch_others() and
    # sch_multiple() group-name keys, which are not actual column names).
    schema_owned <- names(schema$cols)[
        !vapply(
            schema$cols,
            function(col) col$type %in% c("other", "schema_multiple"),
            FALSE
        )
    ]
    # sch_multiple() member columns live in the sch_groups attribute.
    group_owned <- if (!is.null(groups)) unlist(groups, use.names = FALSE) else character(0)
    owned_cols <- union(schema_owned, group_owned)

    changed <- which(old_names != value)
    renamed_owned <- intersect(old_names[changed], owned_cols)

    if (length(renamed_owned) > 0L) {
        cli::cli_abort(
            "Cannot rename schema column{?s} {.field {renamed_owned}} in a {.cls sch_df}.",
            class = "sch_rename_error",
            call = rlang::caller_env()
        )
    }

    NextMethod()
}

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
## usethis namespace: end
NULL

.onLoad <- function(libname, pkgname) {
    rlang::run_on_load() # nocov
}

rlang::on_load(rlang::local_use_cli(format = TRUE, inline = TRUE))

assert <- function(condition, message) {
    x = rlang::enquo(condition)
    if (!isTRUE(rlang::eval_tidy(x))) {
        if (missing(message) || is.null(message)) {
            message = paste("Expected", rlang::expr_deparse(rlang::quo_squash(x))) # nocov
        }
        rlang::abort(message, call = parent.frame(), .envir = parent.frame())
    }
}

#' Define a structured data type
#'
#' Defines the structure of a single 'observation' for a structured data frame.
#' Each column has type restrictions and may be required or optional.
#' Schemas support nesting relationships.
#'
#' @param ... Column specifications, in the form of `col_name = col_type` pairs,
#'   where `col_type` is a call to a column type constructor listed here, such
#'   as `sch_numeric()`. Every type must be a kind of vector, i.e.,
#'   [vctrs::obj_is_vector()] must return `TRUE`.
#'
#'   All columns must be named, except for `sch_others()`, as described below,
#'   and `sch_multiple()`, which describes a group of columns sharing the same
#'   type. A named `sch_nest()` describes columns stored as a nested data frame.
#'
#'   The special function `sch_others()` indicates the preferred location of
#'   other columns not explicitly mentioned in the schema. If no `sch_others()`
#'   appears, then other columns are not allowed.
#'   Trailing commas are permitted.
#' @param .relationships An optional one-sided formula describing the structural
#'   relationships between values in different columns. Formulas can only involve
#'   named arguments to `...`. Use `*` to signify crossed levels, which will
#'   verify all combinations exist, `/` to signify nested levels, and `+` to create
#'   compound keys (bundling columns into a single identifier).
#'   See the examples below.
#' @param desc,.desc A description of the column for consumers of the schema.
#'   The type contraints will be described separately and do not need to be
#'   included in the description.  For example for "age", the description might
#'   be "Age of the patient in years", not "Non-negative integer representing
#'   the age of the patient in years". For the overall `sch_schema`, the `desc`
#'   will be printed as part of the header for data frames implementing the
#'   schema, by default.
#' @param missing If `TRUE`, the column may be contain missing values. Otherwise,
#'   any missing values result in an error.
#' @param required If `TRUE`, the column must be present. If `FALSE`, the column
#'   is optional.
#' @param distinct If `TRUE`, the column must contain no duplicate values (after
#'   accounting for nesting structure).
#'
#' @returns An object of class `sch_schema`,
#'
#' @section Worked examples:
#' Three larger end-to-end example scripts are bundled with the package and
#' show how to build a schema, attach group metadata, validate a compliant
#' data frame, and exercise a range of corruption cases. List them with:
#'
#' ```r
#' list.files(system.file("examples", package = "schmear"), full.names = TRUE)
#' ```
#'
#' The included scripts are `mcmc_draws.R` (fully-crossed MCMC posterior
#' draws), `ei_spec.R` (an ecological-inference specification using
#' [sch_multiple()] groups with a row-sum cross-column check), and
#' `election_data.R` (tidy election returns with a compound primary key and
#' nested/crossed relationship structure).
#'
#' @examples
#' sch_schema(
#'     .desc = "MCMC draws",
#'     .relationships = ~ chain * draw * parameter,
#'     chain = sch_integer("Chain number"),
#'     draw = sch_integer("Draw number", bounds = c(1, Inf), closed = c(TRUE, FALSE)),
#'     parameter = sch_factor("Parameter name", levels = c("mu", "sigma", "log_lik")),
#'     value = sch_numeric("Parameter value")
#' )
#'
#' sch_schema(
#'     .desc = "Student data",
#'     .relationships = ~ (grade + teacher) / table_group,
#'     birthday = sch_date("Date of birth", required = FALSE),
#'     height = sch_numeric(
#'         "Height in inches",
#'         bounds = c(0, 108),
#'         closed = c(FALSE, TRUE)
#'     ),
#'     grade = sch_factor(strict = FALSE, levels = c("Kindergarten", "1st", "2nd")),
#'     teacher = sch_nest(
#'         first = sch_character("First name"),
#'         last = sch_character("Last name")
#'     ),
#'     table_group = sch_integer(bounds=c(1, 6)),
#'     enrolled = sch_logical(missing = FALSE),
#'     sch_others()
#' )
#'
#' sch_schema(
#'     .desc = "Causal inference data",
#'     treatment = sch_factor(levels = c("control", "treatment"), missing = FALSE),
#'     outcome = sch_numeric(missing = FALSE),
#'     sch_multiple("covariates", type = sch_any(missing = FALSE), required = FALSE),
#'     sch_others()
#' )
#'
#' sch_custom(
#'    name = "even",
#'    check = function(x, type) is.integer(x) && all(x %% 2 == 0),
#'    msg = function(type) "vector of even integers",
#'    coerce = function(x, type) (as.integer(x) %/% 2) * 2
#' )
#' @export
sch_schema <- function(..., .desc = NULL, .relationships = NULL) {
    cols = rlang::dots_list(..., .homonyms = "error", .check_assign = TRUE)

    assert(length(cols) > 0, "{.arg ...} must not be empty")
    if (!all(vapply(cols, inherits, FALSE, what = "sch_type"))) {
        rlang::abort("All columns must be specified using a column type constructor.")
    }

    is_other = vapply(cols, function(x) x$type == "other", FALSE)
    is_nest = vapply(cols, function(x) x$type == "schema_nest", FALSE)
    is_multiple = vapply(cols, function(x) x$type == "schema_multiple", FALSE)
    is_unnamed_ok = is_other | is_multiple
    if (sum(is_other) > 1) {
        rlang::abort("Only one {.fn sch_others()} is allowed in a schema.")
    }
    if (any(is_other) && rlang::is_named(cols[is_other])) {
        rlang::abort("{.fn sch_others()} must not be named.")
    }
    if (any(is_multiple) && rlang::is_named(cols[is_multiple])) {
        rlang::abort("{.fn sch_multiple()} must not be named.")
    }
    # Unnamed sch_nest() is no longer supported
    if (any(is_nest & !rlang::have_name(cols))) {
        rlang::abort(
            "Unnamed {.fn sch_nest()} is no longer supported. Use named nests or {.arg .relationships}."
        )
    }

    named_cols = cols[!is_unnamed_ok]
    if (length(named_cols) > 0 && !rlang::is_named(named_cols)) {
        rlang::abort("All columns must be named.")
    }
    nms = names(named_cols)
    if (anyDuplicated(nms[nzchar(nms)]) > 0) {
        rlang::abort("Column names must be unique.")
    }

    # Parse and validate .relationships
    rel_tree = NULL
    if (!is.null(.relationships)) {
        rel_tree = parse_relationship(.relationships)

        # Collect all column names referenced in the formula
        rel_col_names = relationship_columns(rel_tree)

        # All named columns (excluding unnamed sch_others() and sch_multiple())
        regular_names = names(cols)[!is_other & !is_multiple]

        bad_names = setdiff(rel_col_names, regular_names)
        if (length(bad_names) > 0) {
            rlang::abort(c(
                ".relationships references columns not in the schema:",
                paste0("  ", bad_names, collapse = "\n")
            ))
        }

        # Disallow distinct=TRUE columns in formula
        for (cn in rel_col_names) {
            col_type = cols[[cn]]
            if (!is.null(col_type) && isTRUE(attr(col_type, "distinct"))) {
                rlang::abort(c(
                    "Column {.field {cn}} has {.code distinct=TRUE} but appears in {.arg .relationships}.",
                    "i" = "Columns in a relationship formula cannot have {.code distinct=TRUE}."
                ))
            }
        }
    }

    structure(
        list(cols = cols, relationships = rel_tree),
        desc = check_desc(.desc),
        class = "sch_schema"
    )
}


#' @describeIn sch_schema A placeholder for other non-required columns in a schema.
#' @export
sch_others <- function() {
    structure(
        list(type = "other"),
        missing = NA,
        required = FALSE,
        distinct = FALSE,
        class = "sch_type"
    )
}

#' @describeIn sch_schema A column of any type. No type checking is performed.
#' @export
sch_any <- function(desc = NULL, missing = TRUE, required = TRUE, distinct = FALSE) {
    structure(
        list(type = "any"),
        desc = check_desc(desc),
        missing = isTRUE(missing),
        required = isTRUE(required),
        distinct = isTRUE(distinct),
        class = "sch_type"
    )
}

#' @describeIn sch_schema A group of multiple columns sharing the same type. The
#'   group is identified by `name`, which must appear as an entry in the
#'   `sch_groups` attribute of the data frame being validated. That entry is a
#'   character vector of column names that belong to this group.
#'
#'   Optionally accepts cross-column `check`, `msg`, and `coerce` functions that
#'   are applied to the entire group after per-column type checks pass. These
#'   must all be provided together or not at all.
#'
#'   `sch_multiple()` must be unnamed in an `sch_schema()` call. Per-column
#'   constraints such as `missing` and `distinct` are set on the inner `type`
#'   argument.
#'
#' @param name A single string identifying the group. Must match a key in the
#'   `sch_groups` attribute of the data frame.
#' @param type A column type constructor (e.g. [sch_numeric()]) specifying the
#'   expected type of every column in the group.
#' @param required If `TRUE` (default), the group entry in `sch_groups` must
#'   contain at least one column name. If `FALSE`, an empty character vector
#'   for that entry is also accepted.
#' @export
sch_multiple <- function(
    name,
    type,
    desc = NULL,
    required = TRUE,
    check = NULL,
    msg = NULL,
    coerce = NULL
) {
    if (!is.character(name) || length(name) != 1) {
        rlang::abort("{.arg name} must be a single string.")
    }

    invalid_types = c("other", "schema_nest", "schema_multiple")
    if (!inherits(type, "sch_type") || type$type %in% invalid_types) {
        rlang::abort(
            "{.arg type} must be an {.cls sch_type} column type constructor ",
            "(not {.fn sch_others()}, {.fn sch_nest()}, or {.fn sch_multiple()})."
        )
    }

    fns_given = c(!is.null(check), !is.null(msg), !is.null(coerce))
    if (any(fns_given) && !all(fns_given)) {
        rlang::abort(
            "{.arg check}, {.arg msg}, and {.arg coerce} must all be provided together or not at all."
        )
    }

    if (!is.null(check)) {
        if (!is.function(check) || length(formals(check)) != 2) {
            rlang::abort("{.arg check} must be a function with two arguments: `x` and `type`.")
        }
        if (!is.function(msg) || length(formals(msg)) != 1) {
            rlang::abort("{.arg msg} must be a function with one argument: `type`.")
        }
        if (!is.function(coerce) || length(formals(coerce)) != 2) {
            rlang::abort("{.arg coerce} must be a function with two arguments: `x` and `type`.")
        }
    }

    structure(
        list(
            type = "schema_multiple",
            name = name,
            inner = type,
            cross_check = check,
            cross_msg = msg,
            cross_coerce = coerce
        ),
        desc = check_desc(desc),
        required = isTRUE(required),
        class = "sch_type"
    )
}

#' @describeIn sch_schema A set of columns stored as a nested list-column of
#'   data frames. Must be given a name in the outer `sch_schema()`.
#' @export
sch_nest <- function(..., .desc = NULL) {
    cols = rlang::dots_list(..., .homonyms = "error", .check_assign = TRUE)

    if (!all(vapply(cols, inherits, FALSE, what = "sch_type"))) {
        rlang::abort("All columns must be specified using a column type constructor.")
    }
    has_other = any(vapply(cols, function(x) x$type == "other", FALSE))
    if (has_other) {
        rlang::abort("{.fn sch_others()} is not allowed inside {.fn sch_nest()}.")
    }

    if (length(cols) > 0 && !rlang::is_named(cols)) {
        rlang::abort("All columns inside {.fn sch_nest()} must be named.")
    }
    nms = names(cols)
    if (anyDuplicated(nms[nzchar(nms)]) > 0) {
        rlang::abort("Column names must be unique.")
    }

    structure(
        list(type = "schema_nest", cols = cols),
        desc = check_desc(.desc),
        missing = FALSE,
        required = TRUE,
        class = c("sch_schema", "sch_type")
    )
}


# Formula parsing ---------------------------------------------------------

# Parse a relationship formula into a tree
parse_relationship <- function(formula) {
    if (!inherits(formula, "formula")) {
        rlang::abort("{.arg formula} must be a formula.")
    }
    if (length(formula) == 3L) {
        rlang::abort("{.arg formula} must be one-sided (e.g., {.code ~ a * b}).")
    }
    parse_rel_expr(formula[[2L]])
}

# Recursively parse an R expression from the formula RHS
parse_rel_expr <- function(expr) {
    if (is.symbol(expr)) {
        return(list(type = "var", name = as.character(expr)))
    }
    if (is.call(expr)) {
        op = as.character(expr[[1L]])
        if (op == "(") {
            return(parse_rel_expr(expr[[2L]]))
        }
        if (op %in% c("*", ":")) {
            left = parse_rel_expr(expr[[2L]])
            right = parse_rel_expr(expr[[3L]])
            # Flatten nested crosses into one node
            children = c(
                if (left$type == "cross") left$children else list(left),
                if (right$type == "cross") right$children else list(right)
            )
            return(list(type = "cross", children = children))
        }
        if (op == "/") {
            outer = parse_rel_expr(expr[[2L]])
            inner = parse_rel_expr(expr[[3L]])
            return(list(type = "nest", outer = outer, inner = inner))
        }
        if (op == "+") {
            left = parse_rel_expr(expr[[2L]])
            right = parse_rel_expr(expr[[3L]])
            # Flatten nested compounds into one node
            children = c(
                if (left$type == "compound") left$children else list(left),
                if (right$type == "compound") right$children else list(right)
            )
            return(list(type = "compound", children = children))
        }
    }
    rlang::abort(paste0("Unsupported operator in relationship formula: ", deparse(expr)))
}

# Extract all column names from a relationship tree
relationship_columns <- function(tree) {
    switch(
        tree$type,
        var = tree$name,
        cross = unique(unlist(lapply(tree$children, relationship_columns))),
        compound = unique(unlist(lapply(tree$children, relationship_columns))),
        nest = unique(c(relationship_columns(tree$outer), relationship_columns(tree$inner))),
        rlang::abort(paste0("Unknown relationship node type: ", tree$type))
    )
}


#' @describeIn sch_schema A numeric vector that is optionally constrained to be
#'   within a certain range.
#'
#' @param bounds Length-two vector `c(min, max)` specifying the allowed range of values.
#' @param closed Length-two logical vector specifying whether the bounds are
#'   closed (inclusive) or open (exclusive).
#' @export
sch_numeric <- function(
    desc = NULL,
    bounds = c(-Inf, Inf),
    closed = c(TRUE, TRUE),
    missing = TRUE,
    required = TRUE,
    distinct = FALSE
) {
    check_bounds_closed(bounds, closed)

    structure(
        list(type = "numeric", bounds = bounds, closed = closed),
        desc = check_desc(desc),
        missing = isTRUE(missing),
        required = isTRUE(required),
        distinct = isTRUE(distinct),
        class = "sch_type"
    )
}


#' @describeIn sch_schema An integer vector that is optionally constrained to be
#'   within a certain range.
#' @export
sch_integer <- function(
    desc = NULL,
    bounds = c(-Inf, Inf),
    closed = c(TRUE, TRUE),
    missing = TRUE,
    required = TRUE,
    distinct = FALSE
) {
    check_bounds_closed(bounds, closed)

    structure(
        list(type = "integer", bounds = bounds, closed = closed),
        desc = check_desc(desc),
        missing = isTRUE(missing),
        required = isTRUE(required),
        distinct = isTRUE(distinct),
        class = "sch_type"
    )
}

#' @describeIn sch_schema A logical vector.
#' @export
sch_logical <- function(desc = NULL, missing = TRUE, required = TRUE, distinct = FALSE) {
    structure(
        list(type = "logical"),
        desc = check_desc(desc),
        missing = isTRUE(missing),
        required = isTRUE(required),
        distinct = isTRUE(distinct),
        class = "sch_type"
    )
}

#' @describeIn sch_schema A character vector.
#' @export
sch_character <- function(desc = NULL, missing = TRUE, required = TRUE, distinct = FALSE) {
    structure(
        list(type = "character"),
        desc = check_desc(desc),
        missing = isTRUE(missing),
        required = isTRUE(required),
        distinct = isTRUE(distinct),
        class = "sch_type"
    )
}

#' @describeIn sch_schema A factor with specified levels.
#' @param levels A character vector of factor levels, or NULL not enforce specific levels.
#' @param strict If `TRUE`, only factors with the specified levels are accepted.
#'   If `FALSE`, character vectors with the specified levels are also accepted.
#' @export
sch_factor <- function(
    desc = NULL,
    levels = NULL,
    strict = TRUE,
    missing = TRUE,
    required = TRUE,
    distinct = FALSE
) {
    if (!(is.character(levels) || is.null(levels))) {
        rlang::abort("`levels` must be a character vector.")
    }
    structure(
        list(type = "factor", levels = levels, strict = isTRUE(strict)),
        desc = check_desc(desc),
        missing = isTRUE(missing),
        required = isTRUE(required),
        distinct = isTRUE(distinct),
        class = "sch_type"
    )
}

#' @describeIn sch_schema A Date vector that is optionally constrained to be
#'   within a certain range.
#' @export
sch_date <- function(
    desc = NULL,
    bounds = c(as.Date(-Inf), as.Date(Inf)),
    closed = c(FALSE, FALSE),
    missing = TRUE,
    required = TRUE,
    distinct = FALSE
) {
    check_bounds_closed(bounds, closed)

    structure(
        list(type = "date", bounds = bounds, closed = closed),
        desc = check_desc(desc),
        missing = isTRUE(missing),
        required = isTRUE(required),
        distinct = isTRUE(distinct),
        class = "sch_type"
    )
}
#' @describeIn sch_schema A POSIXct vector that is optionally constrained to be
#'   within a certain range.
#' @export
sch_datetime <- function(
    desc = NULL,
    bounds = c(as.POSIXct(-Inf), as.POSIXct(Inf)),
    closed = c(FALSE, FALSE),
    missing = TRUE,
    required = TRUE,
    distinct = FALSE
) {
    check_bounds_closed(bounds, closed)

    structure(
        list(type = "datetime", bounds = bounds, closed = closed),
        desc = check_desc(desc),
        missing = isTRUE(missing),
        required = isTRUE(required),
        distinct = isTRUE(distinct),
        class = "sch_type"
    )
}

#' @describeIn sch_schema A list-column whose elements satisfy `inherits(_, class)`.
#' @param class A character vector of class names.
#'
#' @export
sch_inherits <- function(desc = NULL, class, missing = TRUE, required = TRUE, distinct = FALSE) {
    structure(
        list(type = "inherits", class = as.character(class)),
        desc = check_desc(desc),
        missing = isTRUE(missing),
        required = isTRUE(required),
        distinct = isTRUE(distinct),
        class = "sch_type"
    )
}

#' @describeIn sch_schema A vector satisfying `inherits(_, class)`.
#' @export
sch_list_of <- function(desc = NULL, class, missing = TRUE, required = TRUE, distinct = FALSE) {
    structure(
        list(type = "list_of", class = as.character(class)),
        desc = check_desc(desc),
        missing = isTRUE(missing),
        required = isTRUE(required),
        distinct = isTRUE(distinct),
        class = "sch_type"
    )
}

#' @describeIn sch_schema A custom type defined by user-provided check, type message,
#'   and coercion functions. Additional named values to be stored along with the
#'   type specification may be passed via `...` and will be available to the
#'   check, message, and coercion function as elements of the `type` argument.
#' @param name A name for the custom type.
#' @param check A two-argument function that checks whether an object satisfies
#'   the type. The first argument is the object to check, and the second is the
#'   full type specification.
#' @param msg A one-argument function that generates a descriptive message
#'   about the type when passed the type object itself. Should not end with a period.
#' @param coerce A two-argument function that attempts to coerce an object to the
#'   type. The first argument is the object to coerce, and the second is the full
#'   type specification.
#'
#' @export
sch_custom <- function(
    name,
    desc = NULL,
    check,
    msg,
    coerce,
    ...,
    missing = TRUE,
    required = TRUE,
    distinct = FALSE
) {
    if (!is.character(name) || length(name) != 1) {
        rlang::abort("{.arg name} must be a single string.")
    }
    reserved_nms = c("other", names(type_fns))
    if (name %in% reserved_nms) {
        rlang::abort(
            "{.arg name} must not be one of the reserved type names: {.field {reserved_nms}}."
        )
    }
    err_fn = function(arg) {
        rlang::abort("{.arg {arg}} must be a function with two arguments: `x` and `type`.")
    }
    if (!is.function(check) || length(formals(check)) != 2) {
        err_fn("check")
    }
    if (!is.function(coerce) || length(formals(coerce)) != 2) {
        err_fn("coerce")
    }
    if (!is.function(msg) || length(formals(msg)) != 1) {
        rlang::abort("{.arg msg} must be a function with one argument `type`.")
    }

    extras = rlang::list2(...)
    if (length(extras) > 0 && !rlang::is_named(extras)) {
        rlang::abort("All additional arguments must be named.")
    }

    structure(
        rlang::list2(
            type = "custom",
            name = name,
            check = check,
            msg = msg,
            coerce = coerce,
            !!!extras
        ),
        desc = check_desc(desc),
        missing = isTRUE(missing),
        required = isTRUE(required),
        distinct = isTRUE(distinct),
        class = "sch_type"
    )
}


# these are covered but covr doesn't recognize them
# nocov start
check_num <- function(x, type) {
    switch(
        type$type,
        numeric = is.numeric(x),
        integer = is.integer(x),
        date = inherits(x, "Date"),
        datetime = inherits(x, "POSIXct")
    ) &&
        (if (type$closed[1]) {
            all(x >= type$bounds[1], na.rm = TRUE)
        } else {
            all(x > type$bounds[1], na.rm = TRUE)
        }) &&
        (if (type$closed[2]) {
            all(x <= type$bounds[2], na.rm = TRUE)
        } else {
            all(x < type$bounds[2], na.rm = TRUE)
        })
}
msg_num <- function(type) {
    out = paste0(type$type, " vector")
    if (!all(is.infinite(type$bounds))) {
        out = paste0(
            out,
            " with values in ",
            if (type$closed[1]) "[" else "(",
            type$bounds[1],
            ", ",
            type$bounds[2],
            if (type$closed[2]) "]" else ")"
        )
    }
    out
}
# nocov end

type_fns = list(
    numeric = list(check = check_num, msg = msg_num, coerce = function(x, type) as.numeric(x)),
    integer = list(check = check_num, msg = msg_num, coerce = function(x, type) as.integer(x)),
    date = list(check = check_num, msg = msg_num, coerce = function(x, type) as.Date(x)),
    datetime = list(check = check_num, msg = msg_num, coerce = function(x, type) as.POSIXct(x)),
    logical = list(
        check = function(x, type) is.logical(x),
        msg = function(type) "logical vector",
        coerce = function(x, type) as.logical(x)
    ),

    factor = list(
        check = function(x, type) {
            if (type$strict) {
                is.factor(x) &&
                    (is.null(type$levels) || identical(levels(x), type$levels))
            } else {
                (is.factor(x) || is.character(x)) &&
                    (is.null(type$levels) || all(x[!is.na(x)] %in% type$levels))
            }
        },
        msg = function(type) {
            out = if (type$strict) "factor" else "factor or character"
            if (!is.null(type$levels)) {
                levs = cli::cli_vec(
                    type$levels,
                    list(
                        "vec-trunc" = 10,
                        "vec-last" = ", or "
                    )
                )
                out = cli::format_inline("{out}; one of {.strong {levs}}")
            }
            out
        },
        coerce = function(x, type) {
            if (!is.null(type$levels)) {
                factor(x, levels = type$levels)
            } else {
                as.factor(x)
            }
        }
    ),

    character = list(
        check = function(x, type) is.character(x),
        msg = function(type) "character vector",
        coerce = function(x, type) as.character(x)
    ),

    any = list(
        check = function(x, type) TRUE,
        msg = function(type) "vector of any type",
        coerce = function(x, type) x
    ),

    inherits = list(
        check = function(x, type) {
            inherits(x, type$class)
        },
        msg = function(type) {
            cli::format_inline("vector inheriting from {.cls {type$class}}")
        },
        coerce = function(x, type) {
            methods::as(x, type$class)
        }
    ),
    list_of = list(
        check = function(x, type) {
            is.list(x) &&
                all(vapply(x, function(e) is.null(e) || inherits(e, type$class), logical(1)))
        },
        msg = function(type) {
            cli::format_inline("list-column with elements of type {.cls {type$class}}")
        },
        coerce = function(x, type) {
            lapply(x, function(y) {
                methods::as(y, type$class)
            })
        }
    ),

    custom = list(
        check = function(x, type) {
            type$check(x, type)
        },
        msg = function(type) {
            type$msg(type)
        },
        coerce = function(x, type) {
            type$coerce(x, type)
        }
    )
)


check_bounds_closed = function(bounds, closed) {
    assert(length(bounds) == 2, "{.arg bounds} must be a length-two vector.")
    assert(
        length(closed) == 2 && is.logical(closed),
        "{.arg closed} must be a length-two logical vector."
    )
}
check_desc = function(desc) {
    if (is.null(desc)) {
        NULL
    } else if (is.character(desc) && length(desc) == 1) {
        desc
    } else {
        rlang::abort("{.arg desc} must be NULL or a single string.", call = parent.frame())
    }
}

# Printing -----------

# Internal helper: formats an "other" column type
format_col_other <- function(tt, col_nm, ansi, depth) {
    out = format(tt, ansi = ansi)
    list(out = c(out), nms = c("..."), levels = c(depth))
}

# Internal helper: formats a "schema_nest" column type
format_col_nest <- function(tt, col_nm, ansi, depth) {
    constraint_parts = c(
        if (!attr(tt, "required")) "optional"
    )
    all_parts = c(constraint_parts, "nested")
    mode_label = paste0("(", paste(all_parts, collapse = "; "), ")")
    if (isTRUE(ansi)) {
        mode_label = cli::col_grey(mode_label)
    }
    desc = attr(tt, "desc")
    hdr = if (!is.null(desc)) {
        paste0(mode_label, " ", desc, ":")
    } else {
        paste0(mode_label, ":")
    }
    # Header lives at the current depth; inner columns at depth+1
    out = c(hdr)
    nms = c(col_nm)
    levels = c(depth)
    inner = format_schema_cols(tt$cols, ansi = ansi, depth = depth + 1L)
    list(out = c(out, inner$out), nms = c(nms, inner$nms), levels = c(levels, inner$levels))
}

# Internal helper: formats a "schema_multiple" column type
format_col_multiple <- function(tt, col_nm, ansi, depth) {
    constraint_parts = c(
        if (!attr(tt, "required")) "optional",
        if (attr(tt$inner, "distinct")) "distinct"
    )
    all_parts = c("multiple", constraint_parts)
    mode_label = paste0("(", paste(all_parts, collapse = "; "), ")")
    if (isTRUE(ansi)) {
        mode_label = cli::col_grey(mode_label)
    }
    desc = attr(tt, "desc")
    hdr = if (!is.null(desc)) {
        paste0(mode_label, " ", desc, ": each ")
    } else {
        paste0(mode_label, ": each ")
    }
    inner_fmt = format(tt$inner, ansi = ansi, capitalize = FALSE)
    inner_fmt = paste0(
        inner_fmt,
        ".",
        if (!attr(tt$inner, "missing")) " No NAs allowed."
    )
    if (!is.null(tt$cross_msg)) {
        inner_fmt = paste0(inner_fmt, " Cross-column: ", tt$cross_msg(tt), ".")
    }
    combined_fmt = paste0(hdr, unname(inner_fmt))
    list(out = c(combined_fmt), nms = c(tt$name), levels = c(depth))
}

# Internal helper: formats a regular column type
format_col_regular <- function(tt, col_nm, ansi, depth) {
    fmt = format(tt, ansi = ansi)
    desc_nm = names(fmt)
    fmt = paste0(
        fmt,
        ".",
        if (!attr(tt, "missing")) " No NAs allowed."
    )
    names(fmt) = desc_nm
    if (!is.null(names(fmt))) {
        fmt = paste0(names(fmt), ": ", fmt)
    }
    # Add constraint label at the very start
    constraint_parts = c(
        if (!attr(tt, "required")) "optional",
        if (attr(tt, "distinct")) "distinct"
    )
    if (length(constraint_parts) > 0L) {
        constraint_label = paste0("(", paste(constraint_parts, collapse = "; "), ")")
        if (isTRUE(ansi)) {
            constraint_label = cli::col_grey(constraint_label)
        }
        fmt = paste0(constraint_label, " ", fmt)
    }
    list(out = c(unname(fmt)), nms = c(col_nm), levels = c(depth))
}

# Format a relationship tree as a human-readable string
format_relationship <- function(tree) {
    switch(
        tree$type,
        var = tree$name,
        cross = paste(
            vapply(tree$children, fmt_rel_child, "", parent_type = "cross"),
            collapse = " \u00d7 "
        ),
        compound = {
            inner = paste(vapply(tree$children, format_relationship, ""), collapse = " + ")
            paste0("(", inner, ")")
        },
        nest = {
            outer_str = fmt_rel_child(tree$outer, "nest")
            inner_str = fmt_rel_child(tree$inner, "nest")
            paste0(outer_str, " / ", inner_str)
        },
        rlang::abort(paste0("Unknown relationship node type: ", tree$type))
    )
}

# Format a child node, adding parens when the child type conflicts with the parent type
# (cross child that is a nest node, or nest child that is a cross node)
fmt_rel_child <- function(child, parent_type) {
    s <- format_relationship(child)
    needs_parens <- switch(
        parent_type,
        cross = child$type == "nest",
        nest = child$type == "cross",
        FALSE
    )
    if (needs_parens) paste0("(", s, ")") else s
}

# Internal helper: recursively formats schema columns, tracking nesting depth.
# Returns a list(out, nms, levels).
format_schema_cols <- function(cols, ansi = FALSE, depth = 0L) {
    out = character(0)
    nms = character(0)
    levels = integer(0)

    for (i in seq_along(cols)) {
        tt = cols[[i]]
        col_nm = names(cols)[i]

        col_result = switch(
            tt$type,
            other = format_col_other(tt, col_nm, ansi, depth),
            schema_nest = format_col_nest(tt, col_nm, ansi, depth),
            schema_multiple = format_col_multiple(tt, col_nm, ansi, depth),
            format_col_regular(tt, col_nm, ansi, depth)
        )

        out = c(out, col_result$out)
        nms = c(nms, col_result$nms)
        levels = c(levels, col_result$levels)
    }
    list(out = out, nms = nms, levels = levels)
}

#' @export
format.sch_schema <- function(x, ansi = FALSE, ...) {
    res = format_schema_cols(x$cols, ansi = ansi, depth = 0L)
    out = res$out
    names(out) = res$nms
    attr(out, "desc") = attr(x, "desc")
    out
}

#' @export
print.sch_schema <- function(x, ...) {
    if (!is.null(attr(x, "desc"))) {
        cat(cli::style_bold(attr(x, "desc")), "\n", sep = "")
    }
    n_req = sum(vapply(x$cols, function(y) attr(y, "required"), FALSE))
    hdr = cli::format_inline("A schema with {n_req} required element{?s}:")
    cat(cli::col_grey(hdr), "\n", sep = "")

    l_fmt = format_schema_cols(x$cols, ansi = TRUE, depth = 0L)
    fmt = l_fmt$out
    nms = l_fmt$nms
    lvls = l_fmt$levels
    max_lvl = max(lvls)

    # Per-level max name width (empty-string names contribute 0)
    w_by_lvl = vapply(0:max_lvl, function(l) max(cli::ansi_nchar(nms[lvls == l])), 0L)

    # Cumulative indent at each level: level 0 has no indent;
    # level L is indented by sum of (w_by_lvl[0:L-1] + 2) each
    cum_indent = integer(max_lvl + 1L)
    for (l in seq_len(max_lvl)) {
        cum_indent[l + 1L] = cum_indent[l] + w_by_lvl[l] + 2L
    }

    console_w = cli::console_width()
    for (i in seq_along(fmt)) {
        l = lvls[i]
        indent = strrep(" ", cum_indent[l + 1L])
        lbl = cli::ansi_align(cli::col_green(nms[i]), width = w_by_lvl[l + 1L], align = "right")
        prefix_width = cum_indent[l + 1L] + w_by_lvl[l + 1L] + 2L
        text_width = max(20L, console_w - prefix_width)
        lines = cli::ansi_strwrap(fmt[i], width = text_width)
        cat(indent, lbl, "  ", lines[1L], "\n", sep = "")
        if (length(lines) > 1L) {
            cont_indent = strrep(" ", prefix_width)
            for (j in seq_len(length(lines) - 1L) + 1L) {
                cat(cont_indent, lines[j], "\n", sep = "")
            }
        }
    }

    # Print relationships if present
    if (!is.null(x$relationships)) {
        cat(
            cli::col_grey("Relationships:"),
            " ",
            cli::style_italic(format_relationship(x$relationships)),
            "\n",
            sep = ""
        )
    }
}

#' @export
format.sch_type <- function(x, ansi = FALSE, capitalize = TRUE, ...) {
    if (x$type == "other") {
        if (isTRUE(ansi)) {
            return(cli::style_italic("Other columns"))
        } else {
            return("Other columns")
        }
    }
    if (x$type == "schema_multiple") {
        inner_msg = type_fns[[x$inner$type]]$msg(x$inner)
        if (!isTRUE(ansi)) {
            inner_msg = cli::ansi_strip(inner_msg)
        }
        a_an = if (grepl("^[aeiou]", inner_msg)) "an " else "a "
        out = paste0("(multiple) ", x$name, ": each column is ", a_an, inner_msg)
        names(out) = attr(x, "desc")
        return(out)
    }
    assert(x$type %in% names(type_fns), paste0("Unknown type: ", x$type))
    msg = type_fns[[x$type]]$msg(x)
    if (!isTRUE(ansi)) {
        msg = cli::ansi_strip(msg)
    }
    a_an = if (grepl("^[aeiou]", msg)) "An " else "A "
    if (!capitalize) {
        a_an = tolower(a_an)
    }
    out = paste0(a_an, msg)
    names(out) = attr(x, "desc")
    out
}
#' @export
print.sch_type <- function(x, ...) {
    fmt = format(x, ansi = TRUE)
    if (!is.null(names(fmt))) {
        cat(names(fmt), ": ", fmt, "\n", sep = "")
    } else {
        cat(fmt, "\n", sep = "")
    }
}

skip_if_not_installed("dplyr")
library(dplyr)

# Shared fixtures --------------------------------------------------------

make_df <- function() {
    schema <- sch_schema(
        .desc = "test",
        .relationships = ~ chain * draw,
        chain = sch_integer(),
        draw = sch_integer(),
        value = sch_numeric(missing = FALSE)
    )
    d <- data.frame(
        chain = c(1L, 1L, 2L, 2L),
        draw = c(1L, 2L, 1L, 2L),
        value = c(1.1, 2.2, 3.3, 4.4)
    )
    new_sch_df(d, schema, class = "mcmc_df")
}

make_simple_df <- function() {
    schema <- sch_schema(
        id = sch_integer(distinct = TRUE),
        name = sch_character(missing = FALSE)
    )
    d <- data.frame(id = 1:3L, name = c("a", "b", "c"))
    new_sch_df(d, schema)
}


# dplyr_row_slice ---------------------------------------------------------

test_that("filter() preserves class and schema", {
    df <- make_simple_df()
    out <- filter(df, id > 1L)
    expect_s3_class(out, "sch_df")
    expect_identical(attr(out, "sch_schema"), attr(df, "sch_schema"))
})

test_that("slice() preserves class and schema", {
    df <- make_simple_df()
    out <- slice(df, 1:2)
    expect_s3_class(out, "sch_df")
})

test_that("arrange() preserves class and schema", {
    df <- make_simple_df()
    out <- arrange(df, desc(id))
    expect_s3_class(out, "sch_df")
})

test_that("dplyr_row_slice does not check relationships by default", {
    # Remove a row to break the chain * draw crossing
    df <- make_df()
    expect_no_error(filter(df, !(chain == 1L & draw == 1L)))
})

test_that("dplyr_row_slice checks relationships when flag is set", {
    df <- make_df()
    expect_error(
        dplyr::dplyr_row_slice(df, c(2L, 3L, 4L), .check_relationships = TRUE),
        class = "sch_validation_error"
    )
})

test_that("dplyr_row_slice passes relationships check when satisfied", {
    df <- make_df()
    # All four rows satisfy chain * draw crossing
    expect_no_error(
        dplyr::dplyr_row_slice(df, 1:4, .check_relationships = TRUE)
    )
})


# dplyr_col_modify --------------------------------------------------------

test_that("mutate() adding a valid column succeeds when schema has sch_others()", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_others()
    )
    df <- new_sch_df(data.frame(id = 1:3L), schema)
    expect_no_error(mutate(df, new_col = id * 2L))
})

test_that("mutate() changing a column to the correct type succeeds", {
    df <- make_simple_df()
    expect_no_error(mutate(df, name = toupper(name)))
})

test_that("mutate() changing a column to wrong type errors", {
    df <- make_simple_df()
    expect_error(
        mutate(df, id = as.character(id)),
        class = "sch_validation_error"
    )
})

test_that("mutate() setting required column to NULL (deleting it) errors", {
    df <- make_simple_df()
    expect_error(
        mutate(df, id = NULL),
        class = "sch_validation_error"
    )
})

test_that("mutate() introducing NA into missing=FALSE column errors", {
    df <- make_simple_df()
    expect_error(
        mutate(df, name = if_else(id == 1L, NA_character_, name)),
        class = "sch_validation_error"
    )
})

test_that("mutate() creating duplicates in distinct=TRUE column errors", {
    df <- make_simple_df()
    expect_error(
        mutate(df, id = 1L),
        class = "sch_validation_error"
    )
})

test_that("mutate() adding extra column to strict schema errors", {
    df <- make_simple_df()
    # schema has no sch_others(), so extra columns are not allowed
    expect_error(
        mutate(df, extra = TRUE),
        class = "sch_validation_error"
    )
})


# dplyr_reconstruct -------------------------------------------------------

test_that("inner_join() with schema-preserving result keeps class", {
    df <- make_simple_df()
    lookup <- tibble(id = 1:3L, score = c(10, 20, 30))
    schema <- sch_schema(
        id = sch_integer(distinct = TRUE),
        name = sch_character(missing = FALSE),
        sch_others()
    )
    df2 <- new_sch_df(data.frame(id = 1:3L, name = c("a", "b", "c")), schema)
    expect_no_error({
        out <- inner_join(df2, lookup, by = "id")
        expect_s3_class(out, "sch_df")
    })
})

test_that("left_join() that introduces NA into missing=FALSE column errors", {
    schema <- sch_schema(
        id = sch_integer(),
        name = sch_character(missing = FALSE)
    )
    df <- new_sch_df(data.frame(id = 1:3L, name = c("a", "b", "c")), schema)
    # right-side has no match for id=3, so left_join introduces NA in `name`
    right <- data.frame(id = 1:2L, name = c("x", "y"))
    expect_error(
        left_join(df, right, by = "id", suffix = c("", ".y")),
        class = "sch_validation_error"
    )
})


# [.sch_df ----------------------------------------------------------------

test_that("selecting all columns preserves class and passes validation", {
    df <- make_simple_df()
    out <- select(df, everything())
    expect_s3_class(out, "sch_df")
})

test_that("relocate() preserves all columns and class", {
    df <- make_simple_df()
    out <- relocate(df, name)
    expect_s3_class(out, "sch_df")
    expect_identical(sort(names(out)), sort(names(df)))
})

test_that("select() removing a required column errors", {
    df <- make_simple_df()
    expect_error(
        select(df, name),
        class = "sch_validation_error"
    )
})

test_that("1-d [ preserves schema on full column selection", {
    df <- make_simple_df()
    out <- df[c("id", "name")]
    expect_s3_class(out, "sch_df")
})

test_that("2-d [ (row subsetting) does not trigger names validation", {
    df <- make_simple_df()
    # Row subsetting should not invoke sch_validate at all
    expect_no_error(df[1:2, ])
    expect_s3_class(df[1:2, ], "sch_df")
})


# names<- / rename() / rename_with() -------------------------------------

test_that("rename() of a schema column errors", {
    df <- make_simple_df()
    expect_error(rename(df, new_id = id), class = "sch_rename_error")
})

test_that("rename() error message mentions the column name", {
    df <- make_simple_df()
    expect_error(rename(df, new_id = id), regexp = "id")
})

test_that("rename() of a non-schema column in a permissive schema succeeds", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_others()
    )
    df <- new_sch_df(data.frame(id = 1:3L, aux = letters[1:3]), schema)
    expect_no_error(rename(df, helper = aux))
})

test_that("rename_with() applied to schema columns errors", {
    df <- make_simple_df()
    expect_error(
        rename_with(df, toupper),
        class = "sch_rename_error"
    )
})

test_that("rename_with() applied only to non-schema columns succeeds", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_others()
    )
    df <- new_sch_df(data.frame(id = 1:3L, aux = letters[1:3]), schema)
    expect_no_error(rename_with(df, toupper, .cols = "aux"))
})

test_that("rename_with() applied to schema column subset errors", {
    df <- make_simple_df()
    expect_error(
        rename_with(df, toupper, .cols = "name"),
        class = "sch_rename_error"
    )
})

# sch_validate() ---------------------------------------------------------

test_that("sch_validate() returns data invisibly on valid input", {
    schema <- sch_schema(
        x = sch_numeric(),
        y = sch_character()
    )
    df <- data.frame(x = 1:3, y = c("a", "b", "c"))
    result <- sch_validate(schema, df)
    expect_identical(result, df)
})

test_that("sch_validate() passes valid data with all column types", {
    schema <- sch_schema(
        a = sch_numeric(),
        b = sch_integer(),
        c = sch_logical(),
        d = sch_character(),
        e = sch_factor(levels = c("x", "y")),
        f = sch_date(),
        g = sch_datetime()
    )
    df <- data.frame(
        a = c(1.1, 2.2),
        b = c(1L, 2L),
        c = c(TRUE, FALSE),
        d = c("hello", "world"),
        e = factor(c("x", "y"), levels = c("x", "y")),
        f = as.Date(c("2020-01-01", "2020-01-02")),
        g = as.POSIXct(c("2020-01-01 12:00", "2020-01-02 12:00"))
    )
    expect_no_error(sch_validate(schema, df))
})

test_that("sch_validate() errors on non-schema input", {
    expect_error(sch_validate(list(), data.frame()), "sch_schema")
})

test_that("sch_validate() errors on non-data-frame input", {
    schema <- sch_schema(x = sch_numeric())
    expect_error(sch_validate(schema, list(x = 1)), "data frame")
})

# Required columns -------------------------------------------------------

test_that("missing required column raises error", {
    schema <- sch_schema(
        x = sch_numeric(),
        y = sch_character()
    )
    df <- data.frame(x = 1:3)
    expect_error(
        sch_validate(schema, df),
        class = "sch_validation_error"
    )
    expect_error(sch_validate(schema, df), "Required.*missing")
})

test_that("missing optional column is allowed", {
    schema <- sch_schema(
        x = sch_numeric(),
        y = sch_character(required = FALSE)
    )
    df <- data.frame(x = 1:3)
    expect_no_error(sch_validate(schema, df))
})

# Missing values ---------------------------------------------------------

test_that("NAs in column with missing=FALSE raises error", {
    schema <- sch_schema(
        x = sch_numeric(missing = FALSE)
    )
    df <- data.frame(x = c(1, NA, 3))
    expect_error(
        sch_validate(schema, df),
        class = "sch_validation_error"
    )
    expect_error(sch_validate(schema, df), "missing values")
})

test_that("NAs in column with missing=TRUE are allowed", {
    schema <- sch_schema(
        x = sch_numeric(missing = TRUE)
    )
    df <- data.frame(x = c(1, NA, 3))
    expect_no_error(sch_validate(schema, df))
})

# Type checks ------------------------------------------------------------

test_that("wrong type raises error", {
    schema <- sch_schema(x = sch_integer())
    df <- data.frame(x = c(1.5, 2.5))
    expect_error(
        sch_validate(schema, df),
        class = "sch_validation_error"
    )
    expect_error(sch_validate(schema, df), "wrong type")
})

test_that("numeric bounds violation raises error", {
    schema <- sch_schema(x = sch_numeric(bounds = c(0, 10)))
    df <- data.frame(x = c(5, 15))
    expect_error(sch_validate(schema, df), "wrong type")
})

test_that("factor column passes type check", {
    schema <- sch_schema(x = sch_factor(levels = c("a", "b")))
    df <- data.frame(x = factor(c("a", "b"), levels = c("a", "b")))
    expect_no_error(sch_validate(schema, df))
})

test_that("character for non-strict factor passes", {
    schema <- sch_schema(x = sch_factor(levels = c("a", "b"), strict = FALSE))
    df <- data.frame(x = c("a", "b"), stringsAsFactors = FALSE)
    expect_no_error(sch_validate(schema, df))
})

test_that("character for strict factor fails", {
    schema <- sch_schema(x = sch_factor(levels = c("a", "b"), strict = TRUE))
    df <- data.frame(x = c("a", "b"), stringsAsFactors = FALSE)
    expect_error(sch_validate(schema, df), "wrong type")
})

test_that("date type check works", {
    schema <- sch_schema(x = sch_date())
    df <- data.frame(x = as.Date(c("2020-01-01", "2020-06-15")))
    expect_no_error(sch_validate(schema, df))

    df_bad <- data.frame(x = c("2020-01-01", "2020-06-15"))
    expect_error(sch_validate(schema, df_bad), "wrong type")
})

test_that("inherits type check works", {
    schema <- sch_schema(x = sch_inherits(class = "Date"))
    df <- data.frame(x = as.Date(c("2020-01-01", "2020-06-15")))
    expect_no_error(sch_validate(schema, df))
})

test_that("sch_any() accepts any type", {
    schema <- sch_schema(x = sch_any())
    # Test with numeric
    df_numeric <- data.frame(x = c(1.5, 2.5))
    expect_no_error(sch_validate(schema, df_numeric))
    # Test with character
    df_char <- data.frame(x = c("a", "b"))
    expect_no_error(sch_validate(schema, df_char))
    # Test with logical
    df_logical <- data.frame(x = c(TRUE, FALSE))
    expect_no_error(sch_validate(schema, df_logical))
    # Test with factor
    df_factor <- data.frame(x = factor(c("a", "b")))
    expect_no_error(sch_validate(schema, df_factor))
})

test_that("sch_any() with mixed types passes", {
    schema <- sch_schema(
        a = sch_numeric(),
        b = sch_any(),
        c = sch_any()
    )
    df <- data.frame(
        a = c(1.1, 2.2),
        b = c("x", "y"),
        c = c(100, 200)
    )
    expect_no_error(sch_validate(schema, df))
})

test_that("sch_any() respects missing=FALSE", {
    schema <- sch_schema(x = sch_any(missing = FALSE))
    df <- data.frame(x = c(1, NA))
    expect_error(sch_validate(schema, df), "missing values")
})

test_that("sch_any() respects missing=TRUE", {
    schema <- sch_schema(x = sch_any(missing = TRUE))
    df <- data.frame(x = c(1, NA))
    expect_no_error(sch_validate(schema, df))
})

test_that("sch_any() respects required=FALSE", {
    schema <- sch_schema(x = sch_any(required = FALSE))
    df <- data.frame()
    expect_no_error(sch_validate(schema, df))
})

test_that("sch_any() with distinct=TRUE detects duplicates", {
    schema <- sch_schema(x = sch_any(distinct = TRUE))
    df <- data.frame(x = c(1, 1, 2))
    expect_error(sch_validate(schema, df), "duplicate")
})

test_that("sch_any() with distinct=TRUE allows unique values", {
    schema <- sch_schema(x = sch_any(distinct = TRUE))
    df <- data.frame(x = c(1, 2, 3))
    expect_no_error(sch_validate(schema, df))
})

# Extra columns ----------------------------------------------------------

test_that("extra columns without sch_others() raises error", {
    schema <- sch_schema(x = sch_numeric())
    df <- data.frame(x = 1, y = 2)
    expect_error(
        sch_validate(schema, df),
        class = "sch_validation_error"
    )
    expect_error(sch_validate(schema, df), "Unexpected")
})

test_that("extra columns with sch_others() are allowed", {
    schema <- sch_schema(x = sch_numeric(), sch_others())
    df <- data.frame(x = 1, y = 2, z = 3)
    expect_no_error(sch_validate(schema, df))
})

# Distinct ---------------------------------------------------------------

test_that("duplicate values in distinct column raises error", {
    schema <- sch_schema(x = sch_integer(distinct = TRUE))
    df <- data.frame(x = c(1L, 2L, 1L))
    expect_error(
        sch_validate(schema, df),
        class = "sch_validation_error"
    )
    expect_error(sch_validate(schema, df), "duplicate")
})

test_that("unique values in distinct column pass", {
    schema <- sch_schema(x = sch_integer(distinct = TRUE))
    df <- data.frame(x = c(1L, 2L, 3L))
    expect_no_error(sch_validate(schema, df))
})

test_that("omitting 'distinct' from check skips distinct checks", {
    schema <- sch_schema(x = sch_integer(distinct = TRUE))
    df <- data.frame(x = c(1L, 1L, 1L))
    expect_no_error(sch_validate(schema, df, check = c("names", "types", "relationships")))
})

# Multiple issues --------------------------------------------------------

test_that("multiple issues are collected and reported", {
    schema <- sch_schema(
        x = sch_numeric(missing = FALSE),
        y = sch_integer()
    )
    df <- data.frame(x = c(1, NA), y = c(1.5, 2.5))
    err <- expect_error(
        sch_validate(schema, df),
        class = "sch_validation_error"
    )
    # Should mention both issues
    expect_match(conditionMessage(err), "missing values")
    expect_match(conditionMessage(err), "wrong type")
})

test_that("issues list is attached to the error condition", {
    schema <- sch_schema(x = sch_numeric(), y = sch_integer())
    df <- data.frame(x = "not_numeric", y = 1.5)
    err <- expect_error(
        sch_validate(schema, df),
        class = "sch_validation_error"
    )
    expect_true(!is.null(err$issues))
    expect_true(length(err$issues) >= 2)
})

# Named nest validation --------------------------------------------------

test_that("named nest validates nested data frames", {
    schema <- sch_schema(
        id = sch_integer(),
        details = sch_nest(
            param = sch_character(),
            value = sch_numeric()
        )
    )
    df <- data.frame(id = c(1L, 2L))
    df$details <- list(
        data.frame(param = c("a", "b"), value = c(1.1, 2.2)),
        data.frame(param = c("a", "b"), value = c(3.3, 4.4))
    )
    expect_no_error(sch_validate(schema, df))
})

test_that("named nest detects non-data-frame element", {
    schema <- sch_schema(
        id = sch_integer(),
        details = sch_nest(
            param = sch_character()
        )
    )
    df <- data.frame(id = c(1L, 2L))
    df$details <- list(
        data.frame(param = "a"),
        "not_a_df"
    )
    expect_error(sch_validate(schema, df), "not a data frame")
})

test_that("named nest detects when column is not a list", {
    schema <- sch_schema(
        id = sch_integer(),
        details = sch_nest(
            param = sch_character()
        )
    )
    df <- data.frame(id = 1L, details = "not_a_list")
    expect_error(sch_validate(schema, df), "should be a list")
})

test_that("named nest detects wrong type in inner column", {
    schema <- sch_schema(
        id = sch_integer(),
        details = sch_nest(
            value = sch_integer()
        )
    )
    df <- data.frame(id = 1L)
    df$details <- list(data.frame(value = c(1.5, 2.5)))
    expect_error(sch_validate(schema, df), "wrong type")
})

test_that("missing required named nest column raises error", {
    schema <- sch_schema(
        id = sch_integer(),
        details = sch_nest(
            param = sch_character()
        )
    )
    df <- data.frame(id = 1L)
    expect_error(sch_validate(schema, df), "Required.*missing")
})

test_that("required named nest column must be present", {
    # sch_nest() does not have a `required` parameter in its constructor,
    # so named nests are always required by default. This test verifies that
    # a missing required named nest raises an error.
    schema <- sch_schema(
        id = sch_integer(),
        details = sch_nest(
            param = sch_character()
        )
    )
    df <- data.frame(id = 1L)
    expect_error(sch_validate(schema, df), "Required.*missing")
})

# Empty data frame -------------------------------------------------------

test_that("empty data frame with all required columns passes", {
    schema <- sch_schema(
        x = sch_numeric(),
        y = sch_character()
    )
    df <- data.frame(x = numeric(0), y = character(0))
    expect_no_error(sch_validate(schema, df))
})

test_that("empty data frame missing required column fails", {
    schema <- sch_schema(
        x = sch_numeric(),
        y = sch_character()
    )
    df <- data.frame(x = numeric(0))
    expect_error(sch_validate(schema, df), "missing")
})

# Error class and message quality ----------------------------------------

test_that("error has class sch_validation_error", {
    schema <- sch_schema(x = sch_numeric())
    df <- data.frame(x = "bad")
    err <- expect_error(
        sch_validate(schema, df),
        class = "sch_validation_error"
    )
    expect_s3_class(err, "sch_validation_error")
})

test_that("error message mentions issue count", {
    schema <- sch_schema(
        x = sch_numeric(missing = FALSE),
        y = sch_integer()
    )
    df <- data.frame(x = c(1, NA), y = c(1.5, 2.5))
    expect_error(sch_validate(schema, df), "2 issues")
})

# Regression tests: check="names" detects missing required columns ----

test_that('check="names" detects missing required regular column', {
    schema <- sch_schema(
        x = sch_numeric(),
        y = sch_character()
    )
    df <- data.frame(x = 1)
    expect_error(
        sch_validate(schema, df, check = "names"),
        class = "sch_validation_error"
    )
    expect_error(sch_validate(schema, df, check = "names"), "Required.*missing")
})

test_that('check="names" does not error on absent optional column', {
    schema <- sch_schema(
        x = sch_numeric(),
        y = sch_character(required = FALSE)
    )
    df <- data.frame(x = 1)
    expect_no_error(sch_validate(schema, df, check = "names"))
})

test_that('check="names" detects missing required named nest column', {
    schema <- sch_schema(
        id = sch_integer(),
        details = sch_nest(param = sch_character())
    )
    df <- data.frame(id = 1L)
    expect_error(
        sch_validate(schema, df, check = "names"),
        class = "sch_validation_error"
    )
    expect_error(sch_validate(schema, df, check = "names"), "Required.*missing")
})

# Regression tests: custom type checks with NAs ----

test_that("NA-unaware custom check with NA input raises sch_validation_error", {
    even <- sch_custom(
        name = "even",
        check = function(x, type) is.integer(x) && all(x %% 2 == 0),
        msg = function(type) "vector of even integers",
        coerce = function(x, type) (as.integer(x) %/% 2L) * 2L
    )
    schema <- sch_schema(x = even)
    df <- data.frame(x = c(2L, NA_integer_))
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
})

test_that("custom check that explicitly returns NA raises sch_validation_error", {
    always_na <- sch_custom(
        name = "always_na",
        check = function(x, type) NA,
        msg = function(type) "a type that always returns NA",
        coerce = function(x, type) x
    )
    schema <- sch_schema(x = always_na)
    df <- data.frame(x = 1)
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
})

test_that("valid custom type with no NAs passes without error", {
    even <- sch_custom(
        name = "even2",
        check = function(x, type) is.integer(x) && all(x %% 2L == 0L),
        msg = function(type) "vector of even integers",
        coerce = function(x, type) (as.integer(x) %/% 2L) * 2L
    )
    schema <- sch_schema(x = even)
    df <- data.frame(x = c(2L, 4L, 6L))
    expect_no_error(sch_validate(schema, df))
})

# Regression tests: distinct=TRUE with multiple NAs ----

test_that("distinct=TRUE, missing=TRUE allows two NAs", {
    schema <- sch_schema(x = sch_integer(distinct = TRUE, missing = TRUE))
    df <- data.frame(x = c(1L, 2L, NA_integer_, NA_integer_))
    expect_no_error(sch_validate(schema, df))
})

test_that("distinct=TRUE, missing=FALSE raises 'missing values' for single NA", {
    schema <- sch_schema(x = sch_integer(distinct = TRUE, missing = FALSE))
    df <- data.frame(x = c(1L, 2L, NA_integer_))
    err <- expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_match(conditionMessage(err), "missing values")
})

test_that("distinct=TRUE, missing=TRUE allows one NA among distinct values", {
    schema <- sch_schema(x = sch_integer(distinct = TRUE, missing = TRUE))
    df <- data.frame(x = c(1L, 2L, 3L, NA_integer_))
    expect_no_error(sch_validate(schema, df))
})

test_that("distinct=TRUE, missing=TRUE detects duplicate non-NA values", {
    schema <- sch_schema(x = sch_integer(distinct = TRUE, missing = TRUE))
    df <- data.frame(x = c(1L, 1L, NA_integer_, NA_integer_))
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_error(sch_validate(schema, df), "duplicate")
})

# Regression tests: named nest completeness and inner distinct ----

test_that("named nest: inner distinct=TRUE all satisfied", {
    schema <- sch_schema(
        id = sch_integer(),
        details = sch_nest(
            param = sch_character(distinct = TRUE),
            value = sch_numeric()
        )
    )
    df <- data.frame(id = c(1L, 2L))
    df$details <- list(
        data.frame(param = c("a", "b"), value = c(1.1, 2.2)),
        data.frame(param = c("a", "b"), value = c(3.3, 4.4))
    )
    expect_no_error(sch_validate(schema, df))
})

test_that("named nest: inner distinct=TRUE violation raises error", {
    schema <- sch_schema(
        id = sch_integer(),
        details = sch_nest(
            param = sch_character(distinct = TRUE),
            value = sch_numeric()
        )
    )
    df <- data.frame(id = c(1L, 2L))
    df$details <- list(
        data.frame(param = c("a", "a"), value = c(1.1, 2.2)),
        data.frame(param = c("a", "b"), value = c(3.3, 4.4))
    )
    expect_error(sch_validate(schema, df), "duplicate")
})

test_that("named nest with single element passes", {
    schema <- sch_schema(
        id = sch_integer(),
        details = sch_nest(
            param = sch_character(),
            value = sch_numeric()
        )
    )
    df <- data.frame(id = 1L)
    df$details <- list(data.frame(param = c("a", "b"), value = c(1.1, 2.2)))
    expect_no_error(sch_validate(schema, df))
})

test_that("named nest: inner missing=FALSE rejects NA", {
    schema <- sch_schema(
        id = sch_integer(),
        details = sch_nest(
            value = sch_numeric(missing = FALSE)
        )
    )
    df <- data.frame(id = c(1L, 2L))
    df$details <- list(
        data.frame(value = c(1.0, 2.0)),
        data.frame(value = c(3.0, NA_real_))
    )
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_error(sch_validate(schema, df), "missing values")
})

test_that("named nest: extra inner column without sch_others raises error", {
    schema <- sch_schema(
        id = sch_integer(),
        details = sch_nest(
            param = sch_character()
        )
    )
    df <- data.frame(id = 1L)
    df$details <- list(data.frame(param = "a", extra_col = 99))
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_error(sch_validate(schema, df), "Unexpected")
})

test_that("named nest: zero-row elements pass", {
    schema <- sch_schema(
        id = sch_integer(),
        details = sch_nest(
            param = sch_character(),
            value = sch_numeric()
        )
    )
    df <- data.frame(id = c(1L, 2L))
    df$details <- list(
        data.frame(param = character(0), value = numeric(0)),
        data.frame(param = character(0), value = numeric(0))
    )
    expect_no_error(sch_validate(schema, df))
})

# Multi-level (deep) nesting validation ----------------------------------

test_that("two-level nest: valid data passes", {
    schema <- sch_schema(
        id = sch_integer(),
        level1 = sch_nest(
            name = sch_character(),
            level2 = sch_nest(value = sch_numeric())
        )
    )
    df <- data.frame(id = 1L)
    df$level1 <- list(
        data.frame(name = "x") |>
            (\(d) {
                d$level2 <- list(data.frame(value = 1.5))
                d
            })()
    )
    expect_no_error(sch_validate(schema, df))
})

test_that("two-level nest: type violation in inner nest raises error", {
    schema <- sch_schema(
        id = sch_integer(),
        level1 = sch_nest(
            name = sch_character(),
            level2 = sch_nest(value = sch_numeric())
        )
    )
    df <- data.frame(id = 1L)
    df$level1 <- list(
        data.frame(name = "x") |>
            (\(d) {
                d$level2 <- list(data.frame(value = "not_numeric"))
                d
            })()
    )
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
})

test_that("two-level nest: missing column in inner nest raises error", {
    schema <- sch_schema(
        id = sch_integer(),
        level1 = sch_nest(
            name = sch_character(),
            level2 = sch_nest(value = sch_numeric())
        )
    )
    df <- data.frame(id = 1L)
    # level2 is missing the `value` column
    df$level1 <- list(
        data.frame(name = "x") |>
            (\(d) {
                d$level2 <- list(data.frame(other = 1.5))
                d
            })()
    )
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
})

test_that("two-level nest: distinct violation in inner nest raises error", {
    schema <- sch_schema(
        id = sch_integer(),
        level1 = sch_nest(
            name = sch_character(),
            level2 = sch_nest(value = sch_numeric(distinct = TRUE))
        )
    )
    df <- data.frame(id = 1L)
    df$level1 <- list(
        data.frame(name = "x") |>
            (\(d) {
                d$level2 <- list(data.frame(value = c(1.5, 1.5)))
                d
            })()
    )
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
})

# sch_multiple() validation ----------------------------------------------

test_that("sch_multiple: valid data with sch_groups passes", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "trt", type = sch_numeric(bounds = c(0, 1)))
    )
    df <- data.frame(id = 1:3, trt_a = c(0.2, 0.5, 0.8), trt_b = c(0.8, 0.5, 0.2))
    attr(df, "sch_groups") <- list(trt = c("trt_a", "trt_b"))
    expect_no_error(sch_validate(schema, df))
})

test_that("sch_multiple: missing sch_groups attribute raises error", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "trt", type = sch_numeric())
    )
    df <- data.frame(id = 1L, trt_a = 0.5)
    # No sch_groups attribute
    err <- expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_match(conditionMessage(err), "sch_groups")
})

test_that("sch_multiple: missing group name in sch_groups raises error", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "trt", type = sch_numeric())
    )
    df <- data.frame(id = 1L, trt_a = 0.5)
    attr(df, "sch_groups") <- list(other_group = "trt_a")
    err <- expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_match(conditionMessage(err), "trt")
})

test_that("sch_multiple: empty group with required=TRUE raises error", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "trt", type = sch_numeric(), required = TRUE)
    )
    df <- data.frame(id = 1L)
    attr(df, "sch_groups") <- list(trt = character(0))
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
})

test_that("sch_multiple: empty group with required=FALSE passes", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "trt", type = sch_numeric(), required = FALSE)
    )
    df <- data.frame(id = 1L)
    attr(df, "sch_groups") <- list(trt = character(0))
    expect_no_error(sch_validate(schema, df))
})

test_that("sch_multiple: group column missing from data raises error", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "trt", type = sch_numeric())
    )
    df <- data.frame(id = 1L)
    attr(df, "sch_groups") <- list(trt = c("trt_a", "trt_b"))
    err <- expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_match(conditionMessage(err), "missing")
})

test_that("sch_multiple: wrong type in group column raises error", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "trt", type = sch_numeric())
    )
    df <- data.frame(id = 1:2, trt_a = c("x", "y"), trt_b = c(0.5, 0.6))
    attr(df, "sch_groups") <- list(trt = c("trt_a", "trt_b"))
    err <- expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_match(conditionMessage(err), "wrong type")
})

test_that("sch_multiple: NA in group column with inner missing=FALSE raises error", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "trt", type = sch_numeric(missing = FALSE))
    )
    df <- data.frame(id = 1:2, trt_a = c(0.5, NA), trt_b = c(0.5, 0.5))
    attr(df, "sch_groups") <- list(trt = c("trt_a", "trt_b"))
    err <- expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_match(conditionMessage(err), "missing values")
})

test_that("sch_multiple: cross-column check passes when valid", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(
            name = "trt",
            type = sch_numeric(bounds = c(0, 1)),
            check = function(x, type) {
                row_sums <- Reduce("+", x)
                all(abs(row_sums - 1) < 1e-9)
            },
            msg = function(type) "set of values summing to 1",
            coerce = function(x, type) x
        )
    )
    df <- data.frame(id = 1:2, trt_a = c(0.3, 0.6), trt_b = c(0.7, 0.4))
    attr(df, "sch_groups") <- list(trt = c("trt_a", "trt_b"))
    expect_no_error(sch_validate(schema, df))
})

test_that("sch_multiple: cross-column check failure raises error", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(
            name = "trt",
            type = sch_numeric(bounds = c(0, 1)),
            check = function(x, type) {
                row_sums <- Reduce("+", x)
                all(abs(row_sums - 1) < 1e-9)
            },
            msg = function(type) "set of values summing to 1",
            coerce = function(x, type) x
        )
    )
    df <- data.frame(id = 1:2, trt_a = c(0.3, 0.6), trt_b = c(0.3, 0.4))
    attr(df, "sch_groups") <- list(trt = c("trt_a", "trt_b"))
    err <- expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_match(conditionMessage(err), "trt|cross")
})

test_that("sch_multiple: group columns not flagged as extra", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "trt", type = sch_numeric())
    )
    df <- data.frame(id = 1L, trt_a = 0.5, trt_b = 0.5)
    attr(df, "sch_groups") <- list(trt = c("trt_a", "trt_b"))
    expect_no_error(sch_validate(schema, df))
})

test_that("sch_multiple: multiple groups in one schema all validated", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "trt", type = sch_numeric(bounds = c(0, 1))),
        sch_multiple(name = "lab", type = sch_character())
    )
    df <- data.frame(
        id = 1:2,
        trt_a = c(0.4, 0.6),
        trt_b = c(0.6, 0.4),
        lab_x = c("a", "b"),
        lab_y = c("c", "d")
    )
    attr(df, "sch_groups") <- list(trt = c("trt_a", "trt_b"), lab = c("lab_x", "lab_y"))
    expect_no_error(sch_validate(schema, df))
})

test_that("sch_multiple: check='names' detects missing sch_groups", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "trt", type = sch_numeric())
    )
    df <- data.frame(id = 1L)
    expect_error(
        sch_validate(schema, df, check = "names"),
        class = "sch_validation_error"
    )
})

# .relationships validation: uniqueness ----------------------------------

test_that("relationships: fully crossed data passes uniqueness check", {
    schema <- sch_schema(
        .relationships = ~ a * b,
        a = sch_integer(),
        b = sch_character(),
        value = sch_numeric()
    )
    df <- data.frame(
        a = c(1L, 1L, 2L, 2L),
        b = c("x", "y", "x", "y"),
        value = c(1.0, 2.0, 3.0, 4.0)
    )
    expect_no_error(sch_validate(schema, df))
})

test_that("relationships: duplicate primary key raises error", {
    schema <- sch_schema(
        .relationships = ~ a * b,
        a = sch_integer(),
        b = sch_character(),
        value = sch_numeric()
    )
    df <- data.frame(
        a = c(1L, 1L, 2L, 2L),
        b = c("x", "x", "x", "y"),
        value = c(1.0, 2.0, 3.0, 4.0)
    )
    expect_error(sch_validate(schema, df), "duplicate")
})

# .relationships validation: crossing ------------------------------------

test_that("relationships: incomplete crossing raises error", {
    schema <- sch_schema(
        .relationships = ~ a * b,
        a = sch_integer(),
        b = sch_character(),
        value = sch_numeric()
    )
    # a has {1,2}, b has {"x","y"}, but missing (2,"y")
    df <- data.frame(
        a = c(1L, 1L, 2L),
        b = c("x", "y", "x"),
        value = c(1.0, 2.0, 3.0)
    )
    expect_error(sch_validate(schema, df), "combinations|crossing|incomplete")
})

test_that("relationships: three-way crossing complete data passes", {
    schema <- sch_schema(
        .relationships = ~ a * b * c,
        a = sch_integer(),
        b = sch_character(),
        c = sch_factor(levels = c("p", "q")),
        value = sch_numeric()
    )
    df <- expand.grid(
        a = 1:2,
        b = c("x", "y"),
        c = factor(c("p", "q"), levels = c("p", "q")),
        stringsAsFactors = FALSE
    )
    df$value <- seq_len(nrow(df))
    expect_no_error(sch_validate(schema, df))
})

test_that("relationships: three-way crossing missing combo raises error", {
    schema <- sch_schema(
        .relationships = ~ a * b * c,
        a = sch_integer(),
        b = sch_character(),
        c = sch_factor(levels = c("p", "q")),
        value = sch_numeric()
    )
    df <- expand.grid(
        a = 1:2,
        b = c("x", "y"),
        c = factor(c("p", "q"), levels = c("p", "q")),
        stringsAsFactors = FALSE
    )
    df$value <- seq_len(nrow(df))
    # Remove last row
    df <- df[-nrow(df), ]
    expect_error(sch_validate(schema, df), "combinations|crossing|incomplete")
})

test_that("relationships: intermediate-level crossing is enforced", {
    schema <- sch_schema(
        .relationships = ~ a / (b * c) / (d * e),
        a = sch_integer(),
        b = sch_integer(),
        c = sch_integer(),
        d = sch_integer(),
        e = sch_integer(),
    )
    df <- expand.grid(a = 1:2, b = 1:4, c = 1:4) |>
        subset(b %% 2 == a %% 2 & c %% 2 == a %% 2)
    df <- rbind(cbind(df, d = 1L, e = 1L), cbind(df, d = 1L, e = 2L))
    df <- df[do.call(order, df), ]

    expect_no_error(sch_validate(schema, df))
    # missing b=1, c=1 for a=1
    expect_error(sch_validate(schema, df[-1, ]), "combinations|crossing|incomplete")
    expect_error(sch_validate(schema, df[-1:-2, ]), "combinations|crossing|incomplete")
})

# .relationships validation: nesting ------------------------------------

test_that("relationships: nesting with different inner values per group passes", {
    schema <- sch_schema(
        .relationships = ~ a / b,
        a = sch_integer(),
        b = sch_character(),
        value = sch_numeric()
    )
    # a=1 has b={"x","y"}, a=2 has b={"z"} — nesting allows different b per a
    df <- data.frame(
        a = c(1L, 1L, 2L),
        b = c("x", "y", "z"),
        value = c(1.0, 2.0, 3.0)
    )
    expect_no_error(sch_validate(schema, df))
})

test_that("relationships: nesting still enforces (a,b) uniqueness", {
    schema <- sch_schema(
        .relationships = ~ a / b,
        a = sch_integer(),
        b = sch_character(),
        value = sch_numeric()
    )
    # Duplicate (1, "x")
    df <- data.frame(
        a = c(1L, 1L, 2L),
        b = c("x", "x", "z"),
        value = c(1.0, 2.0, 3.0)
    )
    expect_error(sch_validate(schema, df), "duplicate")
})

# .relationships validation: compound keys --------------------------------

test_that("relationships: compound key with crossing passes", {
    schema <- sch_schema(
        .relationships = ~ (a + b) * c,
        a = sch_integer(),
        b = sch_character(),
        c = sch_factor(levels = c("p", "q")),
        value = sch_numeric()
    )
    # (a,b) combos: (1,"x"), (2,"y"); c: {"p","q"}
    df <- data.frame(
        a = c(1L, 1L, 2L, 2L),
        b = c("x", "x", "y", "y"),
        c = factor(c("p", "q", "p", "q"), levels = c("p", "q")),
        value = c(1.0, 2.0, 3.0, 4.0)
    )
    expect_no_error(sch_validate(schema, df))
})

test_that("relationships: compound key missing crossing combo raises error", {
    schema <- sch_schema(
        .relationships = ~ (a + b) * c,
        a = sch_integer(),
        b = sch_character(),
        c = sch_factor(levels = c("p", "q")),
        value = sch_numeric()
    )
    # Only (1,"x","p") and (2,"y","q") — missing combos
    df <- data.frame(
        a = c(1L, 2L),
        b = c("x", "y"),
        c = factor(c("p", "q"), levels = c("p", "q")),
        value = c(1.0, 2.0)
    )
    expect_error(sch_validate(schema, df), "combinations|crossing|incomplete")
})

# .relationships validation: hierarchical --------------------------------

test_that("relationships: hierarchical a / (b * c) validates completeness within groups", {
    schema <- sch_schema(
        .relationships = ~ a / (b * c),
        a = sch_integer(),
        b = sch_character(),
        c = sch_factor(levels = c("p", "q")),
        value = sch_numeric()
    )
    # a=1: b x c fully crossed; a=2: b x c fully crossed (can differ)
    df <- data.frame(
        a = c(1L, 1L, 1L, 1L, 2L, 2L),
        b = c("x", "x", "y", "y", "z", "z"),
        c = factor(c("p", "q", "p", "q", "p", "q"), levels = c("p", "q")),
        value = 1:6
    )
    expect_no_error(sch_validate(schema, df))
})

test_that("relationships: hierarchical a / (b * c) detects incomplete crossing within group", {
    schema <- sch_schema(
        .relationships = ~ a / (b * c),
        a = sch_integer(),
        b = sch_character(),
        c = sch_factor(levels = c("p", "q")),
        value = sch_numeric()
    )
    # a=1: b={"x","y"}, c={"p","q"}, but missing (y,q)
    df <- data.frame(
        a = c(1L, 1L, 1L, 2L, 2L),
        b = c("x", "x", "y", "z", "z"),
        c = factor(c("p", "q", "p", "p", "q"), levels = c("p", "q")),
        value = 1:5
    )
    expect_error(sch_validate(schema, df), "combinations|crossing|incomplete")
})

# .relationships validation: complex election-style ----------------------

test_that("relationships: complex nested/crossed formula validates", {
    schema <- sch_schema(
        .relationships = ~ (state + contest) / (party * time),
        state = sch_character(),
        contest = sch_character(),
        party = sch_character(),
        time = sch_integer(),
        value = sch_numeric()
    )
    # 2 races: (CA, Gov), (TX, Sen); party: {D, R}; time: {1, 2}
    df <- expand.grid(
        state = c("CA", "TX"),
        contest = c("Gov", "Sen"),
        party = c("D", "R"),
        time = 1:2,
        stringsAsFactors = FALSE
    )
    # Filter to actual race combos
    df <- df[
        (df$state == "CA" & df$contest == "Gov") |
            (df$state == "TX" & df$contest == "Sen"),
    ]
    df$value <- seq_len(nrow(df))
    expect_no_error(sch_validate(schema, df))
})

# .relationships validation: crossing soundness --------------------------
#
# The crossing check computes: actual = nrow(unique(all cols)),
# expected = prod(nrow(unique(child_cols)) for each child).
# These are equal if and only if the data IS the full Cartesian product,
# by a pigeonhole argument: each child-key can appear with at most
# prod(other children unique counts) combinations. So if actual ==
# expected, every combination must be present.

test_that("relationships: 3-way cross with right marginals but latin-square combos raises error", {
    # a={1,2}, b={x,y}, c={p,q}: unique per column = 2 each, expected 2*2*2=8
    # Latin-square arrangement has only 4 unique (a,b,c) triples -- caught
    schema <- sch_schema(
        .relationships = ~ a * b * c,
        a = sch_integer(),
        b = sch_character(),
        c = sch_character(),
        value = sch_numeric()
    )
    df <- data.frame(
        a = c(1L, 1L, 2L, 2L),
        b = c("x", "y", "x", "y"),
        c = c("p", "q", "q", "p"), # Latin square: each row unique, but 4 != 8
        value = 1:4
    )
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
})

test_that("relationships: compound child aliased to single partner raises error", {
    # (a+b) has 3 unique pairs; c has 3 unique values: expected 9
    # But each (a,b) pair only ever appears with ONE c value -> only 3 unique tuples
    schema <- sch_schema(
        .relationships = ~ (a + b) * c,
        a = sch_integer(),
        b = sch_character(),
        c = sch_character(),
        value = sch_numeric()
    )
    df <- data.frame(
        a = c(1L, 2L, 3L),
        b = c("x", "y", "z"),
        c = c("p", "q", "r"),
        value = 1:3
    )
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
})

test_that("relationships: crossing complete within each group passes", {
    schema <- sch_schema(
        .relationships = ~ g / ((a + b) * c),
        g = sch_integer(),
        a = sch_integer(),
        b = sch_character(),
        c = sch_character(),
        value = sch_numeric()
    )
    # Within each g: (a,b) pairs {(1,"x"),(2,"y")} fully crossed with c {p,q}
    df <- data.frame(
        g = c(1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L),
        a = c(1L, 1L, 2L, 2L, 1L, 1L, 2L, 2L),
        b = c("x", "x", "y", "y", "x", "x", "y", "y"),
        c = c("p", "q", "p", "q", "p", "q", "p", "q"),
        value = 1:8
    )
    expect_no_error(sch_validate(schema, df))
})

test_that("relationships: crossing incomplete within a group raises error", {
    # Same schema as above; within g=1: (a,b) pairs aliased to c values (not fully crossed)
    schema <- sch_schema(
        .relationships = ~ g / ((a + b) * c),
        g = sch_integer(),
        a = sch_integer(),
        b = sch_character(),
        c = sch_character(),
        value = sch_numeric()
    )
    df <- data.frame(
        g = c(1L, 1L, 2L, 2L),
        a = c(1L, 2L, 1L, 2L),
        b = c("x", "y", "x", "y"),
        c = c("p", "q", "p", "q"), # each (a,b) paired with only one c per group
        value = 1:4
    )
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
})

# .relationships: new semantics (bottom-up within-group uniqueness) ------
#
# Intended semantics:
#   - `+` (compound): the tuple of columns must be unique within the current
#     group context (globally at the top level; per-group when nested inside /).
#   - `/` (nesting): the inner term is evaluated within each group defined by
#     the outer term.  The inner term itself must be unique within each group.
#   - `*` (crossing): complete Cartesian product within current group context
#     (unchanged).
#
# Key implication: "nesting" allows the same inner value to appear in multiple
# outer groups — that is NOT a violation.  A violation is a duplicate inner
# value *within the same* outer group.
#
# Note: once implemented, the following existing tests will need updating:
#   - "relationships: duplicate primary key raises error" (line ~849): a
#     duplicate row with ~ a * b currently raises "duplicate" (global key
#     check).  Under new semantics there is no global key check; the error
#     becomes "incomplete_crossing" instead.  The regex match on "duplicate"
#     should be relaxed or the description updated.
#   - test-schema.R "sch_schema() warns when .relationships root is + (compound
#     at top level)": the warning should be removed now that top-level + is
#     meaningful.

# compound (+) at top level ----

test_that("relationships: ~ a + b enforces (a,b) combination uniqueness", {
    schema <- sch_schema(
        .relationships = ~ a + b,
        a = sch_integer(),
        b = sch_character(),
        value = sch_numeric()
    )
    # (1,"x"), (1,"y"), (2,"x") — all (a,b) pairs distinct
    df <- data.frame(
        a = c(1L, 1L, 2L),
        b = c("x", "y", "x"),
        value = c(1.0, 2.0, 3.0)
    )
    expect_no_error(sch_validate(schema, df))
})

test_that("relationships: ~ a + b catches duplicate (a,b) pair", {
    schema <- sch_schema(
        .relationships = ~ a + b,
        a = sch_integer(),
        b = sch_character(),
        value = sch_numeric()
    )
    # (1,"x") appears twice
    df <- data.frame(
        a = c(1L, 1L, 2L),
        b = c("x", "x", "y"),
        value = c(1.0, 2.0, 3.0)
    )
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
})

# nesting (/) bottom-level: simple case ----

test_that("relationships: ~ a / b allows same b value across different a groups", {
    # This is the core nesting semantic: b="x" in both a=1 and a=2 is fine —
    # each a group has unique b values internally.
    schema <- sch_schema(
        .relationships = ~ a / b,
        a = sch_integer(),
        b = sch_character(),
        value = sch_numeric()
    )
    df <- data.frame(
        a = c(1L, 1L, 2L, 2L),
        b = c("x", "y", "x", "z"),
        value = c(1.0, 2.0, 3.0, 4.0)
    )
    expect_no_error(sch_validate(schema, df))
})

test_that("relationships: ~ a / b catches duplicate b within an a group", {
    schema <- sch_schema(
        .relationships = ~ a / b,
        a = sch_integer(),
        b = sch_character(),
        value = sch_numeric()
    )
    # b="x" appears twice within a=1
    df <- data.frame(
        a = c(1L, 1L, 2L),
        b = c("x", "x", "z"),
        value = c(1.0, 2.0, 3.0)
    )
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
})

# nesting (/) bottom-level: compound inner ----

test_that("relationships: ~ foo / (bar + baz) allows same (bar,baz) across different foo groups", {
    schema <- sch_schema(
        .relationships = ~ foo / (bar + baz),
        foo = sch_integer(),
        bar = sch_character(),
        baz = sch_character(),
        value = sch_numeric()
    )
    # (bar="a", baz="x") appears in both foo=1 and foo=2 — allowed
    df <- data.frame(
        foo = c(1L, 1L, 2L, 2L),
        bar = c("a", "a", "a", "b"),
        baz = c("x", "y", "x", "x"),
        value = c(1.0, 2.0, 3.0, 4.0)
    )
    expect_no_error(sch_validate(schema, df))
})

test_that("relationships: ~ foo / (bar + baz) catches duplicate (bar,baz) within a foo group", {
    schema <- sch_schema(
        .relationships = ~ foo / (bar + baz),
        foo = sch_integer(),
        bar = sch_character(),
        baz = sch_character(),
        value = sch_numeric()
    )
    # (bar="a", baz="x") appears twice within foo=1
    df <- data.frame(
        foo = c(1L, 1L, 2L),
        bar = c("a", "a", "b"),
        baz = c("x", "x", "y"),
        value = c(1.0, 2.0, 3.0)
    )
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
})

# nesting (/) bottom-level: compound outer ----

test_that("relationships: ~ (foo + bar) / baz allows same baz across different (foo,bar) groups", {
    schema <- sch_schema(
        .relationships = ~ (foo + bar) / baz,
        foo = sch_integer(),
        bar = sch_character(),
        baz = sch_character(),
        value = sch_numeric()
    )
    # baz="p" appears in (foo=1,bar="x") and (foo=1,bar="y") — allowed
    df <- data.frame(
        foo = c(1L, 1L, 2L),
        bar = c("x", "y", "x"),
        baz = c("p", "p", "q"),
        value = c(1.0, 2.0, 3.0)
    )
    expect_no_error(sch_validate(schema, df))
})

test_that("relationships: ~ (foo + bar) / baz catches duplicate baz within a (foo,bar) group", {
    schema <- sch_schema(
        .relationships = ~ (foo + bar) / baz,
        foo = sch_integer(),
        bar = sch_character(),
        baz = sch_character(),
        value = sch_numeric()
    )
    # baz="p" appears twice within (foo=1, bar="x")
    df <- data.frame(
        foo = c(1L, 1L, 2L),
        bar = c("x", "x", "y"),
        baz = c("p", "p", "q"),
        value = c(1.0, 2.0, 3.0)
    )
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
})

# .relationships validation: edge cases ----------------------------------

test_that("relationships: empty data frame with formula passes", {
    schema <- sch_schema(
        .relationships = ~ a * b,
        a = sch_integer(),
        b = sch_character(),
        value = sch_numeric()
    )
    df <- data.frame(a = integer(0), b = character(0), value = numeric(0))
    expect_no_error(sch_validate(schema, df))
})

test_that("relationships: single-row data passes all checks", {
    schema <- sch_schema(
        .relationships = ~ a * b,
        a = sch_integer(),
        b = sch_character(),
        value = sch_numeric()
    )
    df <- data.frame(a = 1L, b = "x", value = 1.0)
    expect_no_error(sch_validate(schema, df))
})

test_that("relationships: skipped when formula column missing from data", {
    schema <- sch_schema(
        .relationships = ~ a * b,
        a = sch_integer(),
        b = sch_character(),
        value = sch_numeric()
    )
    df <- data.frame(a = 1L, value = 1.0)
    # Names check catches it; relationships should not crash
    err <- expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_match(conditionMessage(err), "missing")
})

test_that("relationships: check='relationships' can be run independently", {
    schema <- sch_schema(
        .relationships = ~ a * b,
        a = sch_integer(),
        b = sch_character(),
        value = sch_numeric()
    )
    df <- data.frame(
        a = c(1L, 1L, 2L, 2L),
        b = c("x", "y", "x", "y"),
        value = c(1.0, 2.0, 3.0, 4.0)
    )
    expect_no_error(sch_validate(schema, df, check = "relationships"))
})

test_that("relationships: no formula means check='relationships' is a no-op", {
    schema <- sch_schema(
        a = sch_integer(),
        b = sch_character()
    )
    df <- data.frame(a = 1L, b = "x")
    expect_no_error(sch_validate(schema, df, check = "relationships"))
})

test_that("relationships: NAs in formula columns are treated as distinct values", {
    # NA is treated as a distinct value by vctrs::vec_unique, so a column
    # with NA can cause incomplete-crossing errors even if non-NA values are complete.
    schema <- sch_schema(
        .relationships = ~ a * b,
        a = sch_integer(),
        b = sch_character()
    )
    # Complete crossing for non-NA combos, but NA in a inflates unique count
    df <- data.frame(
        a = c(1L, 1L, 2L, 2L, NA_integer_),
        b = c("x", "y", "x", "y", "x")
    )
    # NA in `a` means 3 unique values of a (1, 2, NA) but only 5 rows,
    # not the 3*2=6 expected for a full crossing — raises an error.
    expect_error(sch_validate(schema, df), "incomplete|combinations|crossing")
})

# sch_multiple with distinct=TRUE -------------------------------------------

test_that("sch_multiple: inner distinct=TRUE detects duplicates in group column", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "vals", type = sch_numeric(distinct = TRUE))
    )
    # vals_a has duplicate 0.5
    df <- data.frame(id = 1:2, vals_a = c(0.5, 0.5), vals_b = c(0.7, 0.8))
    attr(df, "sch_groups") <- list(vals = c("vals_a", "vals_b"))
    err <- expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_match(conditionMessage(err), "duplicate")
})

test_that("sch_multiple: inner distinct=TRUE passes with unique values in group columns", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "vals", type = sch_numeric(distinct = TRUE))
    )
    df <- data.frame(id = 1:3, vals_a = c(0.5, 0.7, 0.9), vals_b = c(0.3, 0.4, 0.6))
    attr(df, "sch_groups") <- list(vals = c("vals_a", "vals_b"))
    expect_no_error(sch_validate(schema, df))
})

test_that("sch_multiple: inner distinct=TRUE with NAs passes (NAs excluded from check)", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "vals", type = sch_numeric(distinct = TRUE, missing = TRUE))
    )
    df <- data.frame(id = 1:2, vals_a = c(0.5, NA), vals_b = c(0.7, NA))
    attr(df, "sch_groups") <- list(vals = c("vals_a", "vals_b"))
    expect_no_error(sch_validate(schema, df))
})

test_that("sch_multiple: inner distinct=TRUE skips check if sch_groups attribute missing", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "vals", type = sch_numeric(distinct = TRUE))
    )
    df <- data.frame(id = 1L, vals_a = 0.5, vals_b = 0.5)
    # No sch_groups attribute — the distinct check within the group should be skipped
    # (but other validations like missing sch_groups will catch the error)
    err <- expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_match(conditionMessage(err), "sch_groups")
})

test_that("sch_multiple: inner distinct=TRUE skips check if group name not in sch_groups", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "vals", type = sch_numeric(distinct = TRUE))
    )
    df <- data.frame(id = 1L, vals_a = 0.5)
    attr(df, "sch_groups") <- list(other_group = "vals_a")
    # Group "vals" is not in sch_groups, so distinct check is skipped
    err <- expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_match(conditionMessage(err), "vals")
})

test_that("sch_multiple: inner distinct=TRUE skips missing group columns", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "vals", type = sch_numeric(distinct = TRUE))
    )
    df <- data.frame(id = 1L, vals_a = 0.5)
    attr(df, "sch_groups") <- list(vals = c("vals_a", "vals_b"))
    # vals_b is in the group but missing from data — should skip and raise error for missing column
    err <- expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_match(conditionMessage(err), "missing")
})

# .relationships with nested columns ------------------------------------

test_that("relationships: named sch_nest() column accepted as part of compound key", {
    schema <- sch_schema(
        .relationships = ~ (grade + teacher) / table_group,
        grade = sch_factor(strict = FALSE, levels = c("Kindergarten", "1st", "2nd")),
        teacher = sch_nest(
            first = sch_character(),
            last = sch_character()
        ),
        table_group = sch_integer()
    )

    smith_k <- data.frame(first = "Jane", last = "Smith")
    jones_1 <- data.frame(first = "Bob", last = "Jones")

    df <- data.frame(
        grade = factor(
            c("Kindergarten", "1st"),
            levels = c("Kindergarten", "1st", "2nd")
        ),
        table_group = c(1L, 2L)
    )
    df$teacher <- list(smith_k, jones_1)

    expect_no_error(sch_validate(schema, df))
})

# sch_coerce() -----------------------------------------------------------

test_that("sch_coerce() returns data invisibly when types already match", {
    schema <- sch_schema(
        x = sch_numeric(),
        y = sch_character()
    )
    df <- data.frame(x = 1.5, y = "hello")
    result <- sch_coerce(schema, df)
    expect_identical(result, df)
})

test_that("sch_coerce() errors on non-schema input", {
    expect_error(sch_coerce(list(), data.frame()), "sch_schema")
})

test_that("sch_coerce() errors on non-data-frame input", {
    schema <- sch_schema(x = sch_numeric())
    expect_error(sch_coerce(schema, list(x = 1)), "data frame")
})

# Per-type coercion ------------------------------------------------------

test_that("sch_coerce() coerces character to numeric", {
    schema <- sch_schema(x = sch_numeric())
    df <- data.frame(x = c("1.1", "2.2", "3.3"), stringsAsFactors = FALSE)
    result <- sch_coerce(schema, df, validate = FALSE)
    expect_equal(result$x, c(1.1, 2.2, 3.3))
    expect_true(is.numeric(result$x))
})

test_that("sch_coerce() coerces character to integer", {
    schema <- sch_schema(x = sch_integer())
    df <- data.frame(x = c("1", "2", "3"), stringsAsFactors = FALSE)
    result <- sch_coerce(schema, df, validate = FALSE)
    expect_equal(result$x, c(1L, 2L, 3L))
    expect_true(is.integer(result$x))
})

test_that("sch_coerce() coerces integer to logical", {
    schema <- sch_schema(x = sch_logical())
    df <- data.frame(x = c(0L, 1L, 1L))
    result <- sch_coerce(schema, df, validate = FALSE)
    expect_equal(result$x, c(FALSE, TRUE, TRUE))
    expect_true(is.logical(result$x))
})

test_that("sch_coerce() coerces integer to character", {
    schema <- sch_schema(x = sch_character())
    df <- data.frame(x = 1:3)
    result <- sch_coerce(schema, df, validate = FALSE)
    expect_equal(result$x, c("1", "2", "3"))
    expect_true(is.character(result$x))
})

test_that("sch_coerce() coerces character to factor (no levels)", {
    schema <- sch_schema(x = sch_factor())
    df <- data.frame(x = c("a", "b", "a"), stringsAsFactors = FALSE)
    result <- sch_coerce(schema, df, validate = FALSE)
    expect_true(is.factor(result$x))
    expect_setequal(levels(result$x), c("a", "b"))
})

test_that("sch_coerce() coerces character to factor with specified levels", {
    schema <- sch_schema(x = sch_factor(levels = c("a", "b", "c")))
    df <- data.frame(x = c("a", "b", "a"), stringsAsFactors = FALSE)
    result <- sch_coerce(schema, df, validate = FALSE)
    expect_true(is.factor(result$x))
    expect_equal(levels(result$x), c("a", "b", "c"))
})

test_that("sch_coerce() coerces character to Date", {
    schema <- sch_schema(x = sch_date())
    df <- data.frame(x = c("2020-01-01", "2021-06-15"), stringsAsFactors = FALSE)
    result <- sch_coerce(schema, df, validate = FALSE)
    expect_true(inherits(result$x, "Date"))
    expect_equal(result$x, as.Date(c("2020-01-01", "2021-06-15")))
})

test_that("sch_coerce() coerces character to POSIXct", {
    schema <- sch_schema(x = sch_datetime())
    df <- data.frame(x = c("2020-01-01 12:00:00"), stringsAsFactors = FALSE)
    result <- sch_coerce(schema, df, validate = FALSE)
    expect_true(inherits(result$x, "POSIXct"))
})

test_that("sch_coerce() leaves sch_any() columns unchanged", {
    schema <- sch_schema(x = sch_any())
    df <- data.frame(x = c(1L, 2L, 3L))
    result <- sch_coerce(schema, df, validate = FALSE)
    expect_identical(result$x, df$x)
})

# Structural cases -------------------------------------------------------

test_that("sch_coerce() skips absent optional columns", {
    schema <- sch_schema(
        x = sch_numeric(),
        y = sch_character(required = FALSE)
    )
    df <- data.frame(x = 1:3)
    expect_no_error(sch_coerce(schema, df, validate = FALSE))
})

test_that("sch_coerce() leaves sch_others() columns untouched", {
    schema <- sch_schema(x = sch_numeric(), sch_others())
    df <- data.frame(x = "1.5", extra = "hello", stringsAsFactors = FALSE)
    result <- sch_coerce(schema, df, validate = FALSE)
    expect_true(is.numeric(result$x))
    expect_equal(result$extra, "hello")
})

test_that("sch_coerce() recursively coerces nested schema columns", {
    schema <- sch_schema(
        group = sch_character(),
        info = sch_nest(
            x = sch_numeric(),
            n = sch_integer()
        )
    )
    df <- data.frame(group = c("A", "B"), stringsAsFactors = FALSE)
    df$info <- list(
        data.frame(x = "1.5", n = "3", stringsAsFactors = FALSE),
        data.frame(x = "2.7", n = "1", stringsAsFactors = FALSE)
    )
    result <- sch_coerce(schema, df, validate = FALSE)
    expect_true(is.numeric(result$info[[1]]$x))
    expect_true(is.integer(result$info[[1]]$n))
    expect_true(is.numeric(result$info[[2]]$x))
    expect_true(is.integer(result$info[[2]]$n))
})

test_that("sch_coerce() coerces sch_multiple() group columns", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "vals", type = sch_numeric())
    )
    df <- data.frame(id = 1L, v1 = "1.1", v2 = "2.2", stringsAsFactors = FALSE)
    attr(df, "sch_groups") <- list(vals = c("v1", "v2"))
    result <- sch_coerce(schema, df, validate = FALSE)
    expect_true(is.numeric(result$v1))
    expect_true(is.numeric(result$v2))
    expect_equal(result$v1, 1.1)
    expect_equal(result$v2, 2.2)
})

test_that("sch_coerce() skips sch_multiple() with missing sch_groups attribute", {
    schema <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "vals", type = sch_numeric(), required = FALSE)
    )
    df <- data.frame(id = 1L)
    expect_no_error(sch_coerce(schema, df, validate = FALSE))
})

# validate flag ----------------------------------------------------------

test_that("sch_coerce() with validate=TRUE runs sch_validate after coercion", {
    schema <- sch_schema(x = sch_numeric(bounds = c(0, 1)))
    df <- data.frame(x = "5.0", stringsAsFactors = FALSE)
    # Coercion succeeds (character -> numeric), but value is out of bounds
    expect_error(sch_coerce(schema, df, validate = TRUE), class = "sch_validation_error")
})

test_that("sch_coerce() with validate=FALSE skips post-coercion validation", {
    schema <- sch_schema(x = sch_numeric(bounds = c(0, 1)))
    df <- data.frame(x = "5.0", stringsAsFactors = FALSE)
    result <- sch_coerce(schema, df, validate = FALSE)
    expect_equal(result$x, 5.0)
})

test_that("sch_coerce() with validate=TRUE passes when coercion fixes the type", {
    schema <- sch_schema(x = sch_integer())
    df <- data.frame(x = c("1", "2", "3"), stringsAsFactors = FALSE)
    expect_no_error(sch_coerce(schema, df, validate = TRUE))
})

# Coercion errors --------------------------------------------------------

test_that("sch_coerce() raises sch_coercion_error when coercion fails", {
    fail_type <- sch_custom(
        name = "fail",
        check = function(x, type) is.character(x),
        msg = function(type) "character",
        coerce = function(x, type) stop("deliberate coercion failure")
    )
    schema <- sch_schema(x = fail_type)
    df <- data.frame(x = 1:3)
    expect_error(sch_coerce(schema, df), class = "sch_coercion_error")
})

test_that("sch_coercion_error collects all failing columns before throwing", {
    fail_type <- sch_custom(
        name = "fail",
        check = function(x, type) is.character(x),
        msg = function(type) "character",
        coerce = function(x, type) stop("deliberate coercion failure")
    )
    schema <- sch_schema(
        x = fail_type,
        y = fail_type
    )
    df <- data.frame(x = 1:2, y = 3:4)
    err <- expect_error(sch_coerce(schema, df), class = "sch_coercion_error")
    expect_length(err$issues, 2)
})

test_that("sch_schema() returns correct class and structure", {
    x <- sch_schema(
        age = sch_integer(),
        name = sch_character()
    )
    expect_s3_class(x, "sch_schema")
    expect_length(x$cols, 2)
    expect_named(x$cols, c("age", "name"))
})

test_that("sch_schema() allows trailing commas", {
    expect_no_error(
        sch_schema(
            age = sch_integer(),
        )
    )
})

test_that("sch_schema() allows sch_others()", {
    x <- sch_schema(a = sch_numeric(), sch_others())
    expect_length(x$cols, 2)
    # sch_others() has no name
    expect_equal(names(x$cols)[2], "")
})

test_that("sch_schema() errors on non-sch_type columns", {
    expect_error(
        sch_schema(a = 1),
        "column type constructor"
    )
})

test_that("sch_schema() errors on multiple sch_others()", {
    expect_error(
        sch_schema(a = sch_numeric(), sch_others(), sch_others()),
        "Only one"
    )
})

test_that("sch_schema() errors on unnamed columns", {
    expect_error(
        sch_schema(sch_numeric()),
        "must be named"
    )
})

test_that("sch_schema() errors on named sch_others()", {
    expect_error(
        sch_schema(a = sch_numeric(), other = sch_others()),
        "must not be named"
    )
})

test_that("sch_schema() errors on duplicate column names", {
    expect_error(
        sch_schema(a = sch_numeric(), a = sch_integer())
    )
})

# sch_others() -----------------------------------------------------------

test_that("sch_others() returns sch_type with type = 'other'", {
    x <- sch_others()
    expect_s3_class(x, "sch_type")
    expect_equal(x$type, "other")
    expect_false(attr(x, "required"))
})

# Column type constructors -----------------------------------------------

test_that("sch_numeric() returns correct structure", {
    x <- sch_numeric("A number", bounds = c(0, 10), closed = c(TRUE, FALSE))
    expect_s3_class(x, "sch_type")
    expect_equal(x$type, "numeric")
    expect_equal(x$bounds, c(0, 10))
    expect_equal(x$closed, c(TRUE, FALSE))
    expect_equal(attr(x, "desc"), "A number")
    expect_true(attr(x, "missing"))
    expect_true(attr(x, "required"))
})

test_that("sch_numeric() defaults: infinite bounds, both closed, missing/required TRUE", {
    x <- sch_numeric()
    expect_equal(x$bounds, c(-Inf, Inf))
    expect_equal(x$closed, c(TRUE, TRUE))
    expect_null(attr(x, "desc"))
    expect_true(attr(x, "missing"))
    expect_true(attr(x, "required"))
})

test_that("sch_integer() returns correct structure", {
    x <- sch_integer("Count", bounds = c(0L, 100L), missing = FALSE, required = FALSE)
    expect_s3_class(x, "sch_type")
    expect_equal(x$type, "integer")
    expect_false(attr(x, "missing"))
    expect_false(attr(x, "required"))
})

test_that("sch_logical() returns correct structure", {
    x <- sch_logical("Flag", missing = FALSE)
    expect_s3_class(x, "sch_type")
    expect_equal(x$type, "logical")
    expect_equal(attr(x, "desc"), "Flag")
    expect_false(attr(x, "missing"))
    expect_true(attr(x, "required"))
})

test_that("sch_character() returns correct structure", {
    x <- sch_character()
    expect_s3_class(x, "sch_type")
    expect_equal(x$type, "character")
})

test_that("sch_any() returns correct structure", {
    x <- sch_any("Any type column", missing = FALSE, distinct = TRUE)
    expect_s3_class(x, "sch_type")
    expect_equal(x$type, "any")
    expect_equal(attr(x, "desc"), "Any type column")
    expect_false(attr(x, "missing"))
    expect_true(attr(x, "required"))
    expect_true(attr(x, "distinct"))
})

test_that("sch_factor() returns correct structure", {
    levs <- c("a", "b", "c")
    x <- sch_factor(levels = levs)
    expect_s3_class(x, "sch_type")
    expect_equal(x$type, "factor")
    expect_equal(x$levels, levs)
})

test_that("sch_factor() errors when levels is not character", {
    expect_error(sch_factor(levels = 1:3), "`levels` must be a character vector")
})

test_that("sch_date() returns correct structure", {
    x <- sch_date("Birthday", required = FALSE)
    expect_s3_class(x, "sch_type")
    expect_equal(x$type, "date")
    expect_false(attr(x, "required"))
})

test_that("sch_datetime() returns correct structure", {
    x <- sch_datetime()
    expect_s3_class(x, "sch_type")
    expect_equal(x$type, "datetime")
    expect_equal(x$closed, c(FALSE, FALSE))
})

test_that("sch_inherits() returns correct structure", {
    x <- sch_inherits(class = "myClass")
    expect_s3_class(x, "sch_type")
    expect_equal(x$type, "inherits")
    expect_equal(x$class, "myClass")
})

test_that("sch_list_of() returns correct structure", {
    x <- sch_list_of(class = "data.frame")
    expect_s3_class(x, "sch_type")
    expect_equal(x$type, "list_of")
    expect_equal(x$class, "data.frame")
})

# Bounds/closed validation -----------------------------------------------

test_that("bounds must be length two", {
    expect_error(sch_numeric(bounds = c(1, 2, 3)), "length-two")
    expect_error(sch_integer(bounds = 1), "length-two")
})

test_that("closed must be a length-two logical vector", {
    expect_error(sch_numeric(closed = c(1, 0)), "length-two logical")
    expect_error(sch_numeric(closed = TRUE), "length-two logical")
})

# sch_custom() -----------------------------------------------------------

test_that("sch_custom() creates a valid custom type", {
    x <- sch_custom(
        name = "mytype",
        check = function(x, type) is.numeric(x),
        msg = function(type) "a custom numeric",
        coerce = function(x, type) as.numeric(x)
    )
    expect_s3_class(x, "sch_type")
    expect_equal(x$type, "custom")
    expect_equal(x$name, "mytype")
    expect_true(is.function(x$check))
    expect_true(is.function(x$msg))
    expect_true(is.function(x$coerce))
})

test_that("sch_custom() errors on reserved names", {
    expect_error(
        sch_custom(
            "numeric",
            check = function(x, type) TRUE,
            msg = function(type) "x",
            coerce = function(x, type) x
        ),
        "reserved"
    )
    expect_error(
        sch_custom(
            "other",
            check = function(x, type) TRUE,
            msg = function(type) "x",
            coerce = function(x, type) x
        ),
        "reserved"
    )
})

test_that("sch_custom() errors when name is not a single string", {
    expect_error(
        sch_custom(
            c("a", "b"),
            check = function(x, type) TRUE,
            msg = function(type) "x",
            coerce = function(x, type) x
        )
    )
})

test_that("sch_custom() errors on wrong function signatures", {
    # check must have 2 args
    expect_error(
        sch_custom(
            "mytype",
            check = function(x) TRUE,
            msg = function(type) "x",
            coerce = function(x, type) x
        ),
        "check"
    )
    # msg must have 1 arg
    expect_error(
        sch_custom(
            "mytype",
            check = function(x, type) TRUE,
            msg = function(type, extra) "x",
            coerce = function(x, type) x
        ),
        "msg"
    )
    # coerce must have 2 args
    expect_error(
        sch_custom(
            "mytype",
            check = function(x, type) TRUE,
            msg = function(type) "x",
            coerce = function(x) x
        ),
        "coerce"
    )
})

test_that("sch_custom() errors on unnamed extra arguments", {
    # description must be explicitly named so the unnamed value lands in ...
    expect_error(
        sch_custom(
            "mytype",
            desc = NULL,
            check = function(x, type) TRUE,
            msg = function(type) "x",
            coerce = function(x, type) x,
            "unnamed_extra"
        ),
        "named"
    )
})

test_that("sch_custom() stores extra named arguments", {
    x <- sch_custom(
        "mytype",
        check = function(x, type) TRUE,
        msg = function(type) "x",
        coerce = function(x, type) x,
        threshold = 42
    )
    expect_equal(x$threshold, 42)
})

# description validation -------------------------------------------------

test_that("description must be NULL or a single string", {
    expect_no_error(sch_numeric(desc = NULL))
    expect_no_error(sch_numeric(desc = "some text"))
    expect_error(sch_numeric(desc = c("a", "b")))
    expect_error(sch_numeric(desc = 123))
})

# format methods ---------------------------------------------------------

test_that("format.sch_type() returns a character string", {
    expect_type(format(sch_numeric()), "character")
    expect_type(format(sch_integer()), "character")
    expect_type(format(sch_logical()), "character")
    expect_type(format(sch_character()), "character")
    expect_type(format(sch_factor(levels = c("a", "b"))), "character")
    expect_type(format(sch_date()), "character")
    expect_type(format(sch_datetime()), "character")
})

test_that("format.sch_type() for sch_others() returns 'Other columns'", {
    expect_equal(format(sch_others()), "Other columns")
})

test_that("format.sch_type() description becomes name attribute", {
    x <- sch_numeric("Height")
    fmt <- format(x)
    expect_equal(names(fmt), "Height")
})

test_that("format.sch_type() with no description has no name", {
    x <- sch_numeric()
    fmt <- format(x)
    expect_null(names(fmt))
})

test_that("format.sch_type() includes range for bounded numeric", {
    x <- sch_numeric(bounds = c(0, 100))
    fmt <- format(x)
    expect_match(fmt, "0")
    expect_match(fmt, "100")
})

test_that("format.sch_schema() returns named character vector", {
    x <- sch_schema(
        age = sch_integer(),
        name = sch_character()
    )
    fmt <- format(x)
    expect_type(fmt, "character")
    expect_named(fmt, c("age", "name"))
})

test_that("format.sch_schema() labels sch_others() as '...'", {
    x <- sch_schema(a = sch_numeric(), sch_others())
    fmt <- format(x)
    expect_true("..." %in% names(fmt))
})

test_that("format.sch_schema() appends optional/no-NA notes", {
    x <- sch_schema(
        a = sch_numeric(missing = FALSE),
        b = sch_numeric(required = FALSE)
    )
    fmt <- format(x)
    expect_match(fmt["a"], "No NAs allowed")
    expect_match(fmt["b"], "optional")
})

test_that("print.sch_schema() runs without error", {
    x <- sch_schema(age = sch_integer(), name = sch_character())
    expect_output(print(x), "schema")
})

test_that("print.sch_type() runs without error", {
    expect_output(print(sch_numeric("Height")), "Height")
    expect_output(print(sch_others()))
})

# sch_nest() (named only — for list-column nested data frames) -----------

test_that("sch_nest() returns sch_schema class with type schema_nest", {
    x <- sch_nest(
        param = sch_character(),
        value = sch_numeric()
    )
    expect_s3_class(x, "sch_schema")
    expect_s3_class(x, "sch_type")
    expect_equal(x$type, "schema_nest")
})

test_that("sch_nest() stores columns correctly", {
    x <- sch_nest(
        a = sch_integer(),
        b = sch_numeric()
    )
    expect_named(x$cols, c("a", "b"))
    expect_s3_class(x$cols$a, "sch_type")
    expect_s3_class(x$cols$b, "sch_type")
})

test_that("sch_nest() accepts .desc", {
    x <- sch_nest(a = sch_integer(), .desc = "My group")
    expect_equal(attr(x, "desc"), "My group")
})

test_that("sch_nest() errors on non-sch_type columns", {
    expect_error(
        sch_nest(a = 1),
        "column type constructor"
    )
})

test_that("sch_nest() errors on unnamed columns", {
    expect_error(
        sch_nest(sch_integer()),
        "named"
    )
})

test_that("sch_nest() errors when sch_others() is included", {
    expect_error(
        sch_nest(a = sch_integer(), sch_others()),
        "sch_others"
    )
})

test_that("sch_nest() errors on duplicate column names", {
    expect_error(
        sch_nest(a = sch_integer(), a = sch_numeric())
    )
})

# sch_nest() inside sch_schema() -----------------------------------------

test_that("named sch_nest() works in sch_schema()", {
    x <- sch_schema(
        id = sch_integer(),
        draws = sch_nest(
            param = sch_character(),
            value = sch_numeric()
        )
    )
    expect_s3_class(x, "sch_schema")
    expect_equal(x$cols$draws$type, "schema_nest")
})

test_that("unnamed sch_nest() is not allowed in sch_schema()", {
    expect_error(
        sch_schema(
            id = sch_integer(),
            sch_nest(
                param = sch_character(),
                value = sch_numeric()
            )
        ),
        "named"
    )
})

# Print/format with sch_nest() ------------------------------------------

test_that("format.sch_schema() with named sch_nest() shows (nested) header", {
    x <- sch_schema(
        age = sch_integer(),
        scores = sch_nest(
            subject = sch_character(),
            score = sch_numeric(),
            .desc = "exam scores"
        )
    )
    fmt <- format(x)
    header_line <- fmt[grep("nested", fmt)]
    expect_length(header_line, 1)
    expect_match(header_line, "exam scores")
    expect_match(header_line, "\\(nested\\)")
    # named nest has the column name in names(fmt)
    expect_true("scores" %in% names(fmt))
})

test_that("print.sch_schema() with named sch_nest() runs without error", {
    x <- sch_schema(
        id = sch_integer(),
        draws = sch_nest(
            param = sch_character(),
            value = sch_numeric(),
            .desc = "MCMC draws"
        )
    )
    expect_output(print(x), "schema")
})

# Distinct argument tests -----------------------------------------------

test_that("All column types default distinct to FALSE", {
    expect_false(attr(sch_numeric(), "distinct"))
    expect_false(attr(sch_integer(), "distinct"))
    expect_false(attr(sch_logical(), "distinct"))
    expect_false(attr(sch_character(), "distinct"))
    expect_false(attr(sch_factor(levels = c("a", "b")), "distinct"))
    expect_false(attr(sch_date(), "distinct"))
    expect_false(attr(sch_datetime(), "distinct"))
    expect_false(attr(sch_inherits(class = "myClass"), "distinct"))
    expect_false(attr(sch_list_of(class = "data.frame"), "distinct"))
    expect_false(attr(
        sch_custom(
            name = "test",
            check = function(x, type) TRUE,
            msg = function(type) "test",
            coerce = function(x, type) x
        ),
        "distinct"
    ))
    expect_false(attr(sch_others(), "distinct"))
})

test_that("distinct = TRUE is stored as attribute", {
    x <- sch_numeric(distinct = TRUE)
    expect_true(attr(x, "distinct"))

    y <- sch_character(distinct = TRUE)
    expect_true(attr(y, "distinct"))

    z <- sch_factor(levels = c("a", "b"), distinct = TRUE)
    expect_true(attr(z, "distinct"))
})

test_that("distinct = FALSE explicitly is stored as attribute", {
    x <- sch_numeric(distinct = FALSE)
    expect_false(attr(x, "distinct"))
})

test_that("[Distinct] flag appears in schema print output for distinct columns", {
    x <- sch_schema(
        id = sch_integer(distinct = TRUE),
        name = sch_character(),
        email = sch_character(distinct = TRUE)
    )
    output <- capture.output(print(x))
    output_text <- paste(output, collapse = "\n")

    # Check that (distinct) flag appears in output
    expect_match(output_text, "\\(distinct\\)")
})

test_that("[Distinct] flag does not appear for non-distinct columns", {
    x <- sch_schema(
        id = sch_integer(),
        name = sch_character()
    )
    output <- capture.output(print(x))
    output_text <- paste(output, collapse = "\n")

    # (distinct) flag should not appear
    expect_false(grepl("\\(distinct\\)", output_text))
})

test_that("[Distinct] and [Optional] flags work together", {
    x <- sch_schema(
        id = sch_integer(distinct = TRUE, required = FALSE),
        name = sch_character(distinct = TRUE)
    )
    output <- capture.output(print(x))
    output_text <- paste(output, collapse = "\n")

    # Both flags should appear (may be combined as "(optional; distinct)")
    expect_match(output_text, "distinct")
    expect_match(output_text, "optional")
})

test_that("Nested schema columns show [Distinct] flag", {
    x <- sch_schema(
        id = sch_integer(),
        details = sch_nest(
            name = sch_character(distinct = TRUE),
            value = sch_numeric()
        )
    )
    output <- capture.output(print(x))
    output_text <- paste(output, collapse = "\n")

    expect_match(output_text, "\\(distinct\\)")
})

# Regression tests: strict factor level validation ----

test_that("strict factor with wrong levels attribute raises error", {
    schema <- sch_schema(x = sch_factor(levels = c("a", "b"), strict = TRUE))
    df <- data.frame(x = factor(c("x", "y"), levels = c("x", "y")))
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
})

test_that("strict factor with extra levels in attribute raises error", {
    schema <- sch_schema(x = sch_factor(levels = c("a", "b"), strict = TRUE))
    df <- data.frame(x = factor(c("a", "b"), levels = c("a", "b", "c")))
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
})

test_that("strict factor with exact schema levels passes", {
    schema <- sch_schema(x = sch_factor(levels = c("a", "b"), strict = TRUE))
    df <- data.frame(x = factor(c("a", "b"), levels = c("a", "b")))
    expect_no_error(sch_validate(schema, df))
})

test_that("strict factor with values outside schema levels raises error", {
    schema <- sch_schema(x = sch_factor(levels = c("a", "b"), strict = TRUE))
    df <- data.frame(x = factor(c("a", "z"), levels = c("a", "b", "z")))
    expect_error(sch_validate(schema, df), class = "sch_validation_error")
})

test_that("non-strict factor ignores extra entries in levels attribute", {
    schema <- sch_schema(x = sch_factor(levels = c("a", "b"), strict = FALSE))
    df <- data.frame(x = factor(c("a", "b"), levels = c("a", "b", "extra")))
    expect_no_error(sch_validate(schema, df))
})

# Regression tests: non-strict factor with NAs ----

test_that("non-strict factor with character and NA allowed when missing=TRUE", {
    schema <- sch_schema(
        x = sch_factor(levels = c("a", "b"), strict = FALSE, missing = TRUE)
    )
    df <- data.frame(x = c("a", NA), stringsAsFactors = FALSE)
    expect_no_error(sch_validate(schema, df))
})

test_that("non-strict factor with character and NA raises 'missing values' when missing=FALSE", {
    schema <- sch_schema(
        x = sch_factor(levels = c("a", "b"), strict = FALSE, missing = FALSE)
    )
    df <- data.frame(x = c("a", NA), stringsAsFactors = FALSE)
    err <- expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_match(conditionMessage(err), "missing values")
})

test_that("non-strict factor with factor input and NA allowed when missing=TRUE", {
    schema <- sch_schema(
        x = sch_factor(levels = c("a", "b"), strict = FALSE, missing = TRUE)
    )
    df <- data.frame(x = factor(c("a", NA), levels = c("a", "b")))
    expect_no_error(sch_validate(schema, df))
})

# Regression tests: sch_list_of with NULL elements ----

test_that("list_of with NULL element and missing=TRUE passes", {
    schema <- sch_schema(x = sch_list_of(class = "data.frame", missing = TRUE))
    df <- data.frame(x = I(list(data.frame(a = 1), NULL)))
    expect_no_error(sch_validate(schema, df))
})

test_that("list_of with NULL element and missing=FALSE raises error", {
    schema <- sch_schema(x = sch_list_of(class = "data.frame", missing = FALSE))
    df <- data.frame(x = I(list(data.frame(a = 1), NULL)))
    err <- expect_error(sch_validate(schema, df), class = "sch_validation_error")
    expect_match(conditionMessage(err), "missing values")
})

test_that("list_of with all valid elements passes", {
    schema <- sch_schema(x = sch_list_of(class = "data.frame"))
    df <- data.frame(x = I(list(data.frame(a = 1), data.frame(a = 2))))
    expect_no_error(sch_validate(schema, df))
})

# sch_multiple() ---------------------------------------------------------

test_that("sch_multiple() returns sch_type with type schema_multiple", {
    x <- sch_multiple(name = "trt", type = sch_numeric())
    expect_s3_class(x, "sch_type")
    expect_equal(x$type, "schema_multiple")
})

test_that("sch_multiple() stores name, inner type, and required", {
    inner <- sch_numeric("A share", bounds = c(0, 1))
    x <- sch_multiple(name = "demo", type = inner, desc = "Demographics")
    expect_equal(x$name, "demo")
    expect_identical(x$inner, inner)
    expect_true(attr(x, "required"))
    expect_equal(attr(x, "desc"), "Demographics")
})

test_that("sch_multiple() required=FALSE is stored correctly", {
    x <- sch_multiple(name = "grp", type = sch_integer(), required = FALSE)
    expect_false(attr(x, "required"))
})

test_that("sch_multiple() without cross-column fns stores NULLs", {
    x <- sch_multiple(name = "grp", type = sch_integer())
    expect_null(x$cross_check)
    expect_null(x$cross_msg)
    expect_null(x$cross_coerce)
})

test_that("sch_multiple() stores cross-column functions", {
    chk <- function(x, type) TRUE
    msg <- function(type) "ok"
    crc <- function(x, type) x
    x <- sch_multiple(name = "grp", type = sch_numeric(), check = chk, msg = msg, coerce = crc)
    expect_identical(x$cross_check, chk)
    expect_identical(x$cross_msg, msg)
    expect_identical(x$cross_coerce, crc)
})

test_that("sch_multiple() errors when name is missing", {
    expect_error(sch_multiple(type = sch_numeric()), "name")
})

test_that("sch_multiple() errors when name is not a single string", {
    expect_error(sch_multiple(name = c("a", "b"), type = sch_numeric()), "name")
    expect_error(sch_multiple(name = 1L, type = sch_numeric()), "name")
})

test_that("sch_multiple() errors when type is not sch_type", {
    expect_error(sch_multiple(name = "grp", type = list()), "sch_type")
    expect_error(sch_multiple(name = "grp", type = "numeric"), "sch_type")
})

test_that("sch_multiple() errors when type is sch_others()", {
    expect_error(sch_multiple(name = "grp", type = sch_others()), "sch_type")
})

test_that("sch_multiple() errors when type is a schema_nest", {
    expect_error(
        sch_multiple(name = "grp", type = sch_nest(a = sch_numeric())),
        "sch_type"
    )
})

test_that("sch_multiple() errors when only some of check/msg/coerce are provided", {
    expect_error(
        sch_multiple(
            name = "grp",
            type = sch_numeric(),
            check = function(x, type) TRUE
        ),
        "check.*msg.*coerce|all.*provided"
    )
    expect_error(
        sch_multiple(
            name = "grp",
            type = sch_numeric(),
            msg = function(type) "x"
        ),
        "check.*msg.*coerce|all.*provided"
    )
})

test_that("sch_multiple() errors on wrong check arity (must have 2 args)", {
    expect_error(
        sch_multiple(
            name = "grp",
            type = sch_numeric(),
            check = function(x) TRUE,
            msg = function(type) "x",
            coerce = function(x, type) x
        ),
        "check"
    )
})

test_that("sch_multiple() errors on wrong msg arity (must have 1 arg)", {
    expect_error(
        sch_multiple(
            name = "grp",
            type = sch_numeric(),
            check = function(x, type) TRUE,
            msg = function(type, extra) "x",
            coerce = function(x, type) x
        ),
        "msg"
    )
})

test_that("sch_multiple() errors on wrong coerce arity (must have 2 args)", {
    expect_error(
        sch_multiple(
            name = "grp",
            type = sch_numeric(),
            check = function(x, type) TRUE,
            msg = function(type) "x",
            coerce = function(x) x
        ),
        "coerce"
    )
})

# sch_multiple() inside sch_schema() ------------------------------------

test_that("sch_schema() allows unnamed sch_multiple()", {
    x <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "trt", type = sch_numeric())
    )
    expect_s3_class(x, "sch_schema")
    expect_length(x$cols, 2)
    expect_equal(x$cols[[2]]$type, "schema_multiple")
})

test_that("sch_schema() errors on named sch_multiple()", {
    expect_error(
        sch_schema(
            id = sch_integer(),
            grp = sch_multiple(name = "trt", type = sch_numeric())
        ),
        "must not be named"
    )
})

test_that("sch_schema() allows sch_multiple() alongside sch_others()", {
    x <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "trt", type = sch_numeric()),
        sch_others()
    )
    expect_length(x$cols, 3)
})

# format/print for sch_multiple() ---------------------------------------

test_that("format.sch_type() handles schema_multiple", {
    x <- sch_multiple(name = "trt", type = sch_numeric(bounds = c(0, 1)))
    fmt <- format(x)
    expect_type(fmt, "character")
    expect_match(fmt, "trt")
})

test_that("format.sch_schema() includes multiple group entry", {
    x <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "trt", type = sch_numeric(), desc = "Treatments")
    )
    fmt <- format(x)
    # Should have an entry for the multiple group
    expect_true(any(grepl("trt|multiple|Treatments", fmt, ignore.case = TRUE)))
})

test_that("print.sch_schema() runs without error when sch_multiple() present", {
    x <- sch_schema(
        id = sch_integer(),
        sch_multiple(name = "trt", type = sch_numeric())
    )
    expect_output(print(x))
})

# .relationships formula: parsing ----------------------------------------

test_that("parse_relationship() parses ~ a into Var", {
    tree <- parse_relationship(~a)
    expect_equal(tree$type, "var")
    expect_equal(tree$name, "a")
})

test_that("parse_relationship() parses ~ a * b into Cross", {
    tree <- parse_relationship(~ a * b)
    expect_equal(tree$type, "cross")
    expect_length(tree$children, 2)
    expect_equal(tree$children[[1]]$name, "a")
    expect_equal(tree$children[[2]]$name, "b")
})

test_that("parse_relationship() flattens ~ a * b * c into a single Cross", {
    tree <- parse_relationship(~ a * b * c)
    expect_equal(tree$type, "cross")
    expect_length(tree$children, 3)
    expect_equal(tree$children[[1]]$name, "a")
    expect_equal(tree$children[[2]]$name, "b")
    expect_equal(tree$children[[3]]$name, "c")
})

test_that("parse_relationship() parses ~ a / b into Nest", {
    tree <- parse_relationship(~ a / b)
    expect_equal(tree$type, "nest")
    expect_equal(tree$outer$name, "a")
    expect_equal(tree$inner$name, "b")
})

test_that("parse_relationship() parses ~ a / b / c as left-associative nesting", {
    tree <- parse_relationship(~ a / b / c)
    expect_equal(tree$type, "nest")
    expect_equal(tree$outer$type, "nest")
    expect_equal(tree$outer$outer$name, "a")
    expect_equal(tree$outer$inner$name, "b")
    expect_equal(tree$inner$name, "c")
})

test_that("parse_relationship() parses ~ a + b into Compound", {
    tree <- parse_relationship(~ a + b)
    expect_equal(tree$type, "compound")
    expect_length(tree$children, 2)
})

test_that("parse_relationship() flattens ~ a + b + c into a single Compound", {
    tree <- parse_relationship(~ a + b + c)
    expect_equal(tree$type, "compound")
    expect_length(tree$children, 3)
})

test_that("parse_relationship() parses ~ (a + b) * c correctly", {
    tree <- parse_relationship(~ (a + b) * c)
    expect_equal(tree$type, "cross")
    expect_equal(tree$children[[1]]$type, "compound")
    expect_equal(tree$children[[2]]$name, "c")
})

test_that("parse_relationship() parses ~ a * (b / c) correctly", {
    tree <- parse_relationship(~ a * (b / c))
    expect_equal(tree$type, "cross")
    expect_equal(tree$children[[1]]$name, "a")
    expect_equal(tree$children[[2]]$type, "nest")
    expect_equal(tree$children[[2]]$outer$name, "b")
    expect_equal(tree$children[[2]]$inner$name, "c")
})

test_that("parse_relationship() parses ~ (a * b) / c correctly", {
    tree <- parse_relationship(~ (a * b) / c)
    expect_equal(tree$type, "nest")
    expect_equal(tree$outer$type, "cross")
    expect_equal(tree$inner$name, "c")
})

test_that("parse_relationship() treats : as * (synonym)", {
    tree <- parse_relationship(~ a:b)
    expect_equal(tree$type, "cross")
    expect_length(tree$children, 2)
})

test_that("parse_relationship() precedence: * and / are left-associative", {
    # ~ a * b / c  parses as  ~ (a * b) / c  in R
    tree <- parse_relationship(~ a * b / c)
    expect_equal(tree$type, "nest")
    expect_equal(tree$outer$type, "cross")
    expect_equal(tree$inner$name, "c")
})

test_that("parse_relationship() precedence: * binds tighter than +", {
    # ~ a + b * c  should parse as  ~ a + (b * c)
    tree <- parse_relationship(~ a + b * c)
    expect_equal(tree$type, "compound")
    expect_equal(tree$children[[1]]$name, "a")
    expect_equal(tree$children[[2]]$type, "cross")
})

test_that("parse_relationship() complex election-style formula", {
    tree <- parse_relationship(
        ~ (state + election + contest) / ((party / candidate) * time * (geo / method))
    )
    expect_equal(tree$type, "nest")
    expect_equal(tree$outer$type, "compound")
    expect_length(tree$outer$children, 3) # state, election, contest
    expect_equal(tree$inner$type, "cross")
    expect_length(tree$inner$children, 3) # (party/cand), time, (geo/method)
})

test_that("parse_relationship() errors on non-formula input", {
    expect_error(parse_relationship("not a formula"), "formula")
})

test_that("parse_relationship() errors on two-sided formula", {
    expect_error(parse_relationship(y ~ x), "one-sided")
})

# .relationships in sch_schema() -----------------------------------------

test_that("sch_schema() accepts .relationships formula", {
    x <- sch_schema(
        .relationships = ~ a * b,
        a = sch_integer(),
        b = sch_character(),
        value = sch_numeric()
    )
    expect_s3_class(x, "sch_schema")
    expect_false(is.null(x$relationships))
    expect_equal(x$relationships$type, "cross")
})

test_that("sch_schema() with .relationships = NULL stores no relationships", {
    x <- sch_schema(a = sch_integer())
    expect_null(x$relationships)
})

test_that("sch_schema() allows a named sch_nest() column in .relationships", {
    expect_no_error(sch_schema(
        .relationships = ~ (grade + teacher) / table_group,
        grade = sch_factor(strict = FALSE, levels = c("Kindergarten", "1st", "2nd")),
        teacher = sch_nest(
            first = sch_character(),
            last = sch_character()
        ),
        table_group = sch_integer()
    ))
})

test_that("sch_schema() errors when .relationships references non-existent column", {
    expect_error(
        sch_schema(
            .relationships = ~ a * b,
            a = sch_integer()
        ),
        "b"
    )
})

test_that("sch_schema() errors when .relationships references distinct=TRUE column", {
    expect_error(
        sch_schema(
            .relationships = ~ a * b,
            a = sch_integer(distinct = TRUE),
            b = sch_character()
        ),
        "distinct"
    )
})

test_that("sch_schema() errors when .relationships is not a formula", {
    expect_error(
        sch_schema(
            .relationships = "not a formula",
            a = sch_integer()
        ),
        "formula"
    )
})

test_that("sch_schema() does not warn when .relationships root is + (compound at top level)", {
    expect_no_warning(
        sch_schema(
            .relationships = ~ a + b,
            a = sch_integer(),
            b = sch_character()
        )
    )
})

# format/print with .relationships ---------------------------------------

test_that("print.sch_schema() shows relationships when present", {
    x <- sch_schema(
        .relationships = ~ a * b,
        a = sch_integer(),
        b = sch_character(),
        value = sch_numeric()
    )
    output <- capture.output(print(x))
    output_text <- paste(output, collapse = "\n")
    expect_match(output_text, "Relationships|relationships")
})

test_that("format.sch_schema() includes formula columns in flat list", {
    x <- sch_schema(
        .relationships = ~ chain * draw * param,
        chain = sch_integer(),
        draw = sch_integer(),
        param = sch_factor(levels = c("mu", "sigma")),
        value = sch_numeric()
    )
    fmt <- format(x)
    # All columns should be present in the flat list
    expect_true("chain" %in% names(fmt))
    expect_true("draw" %in% names(fmt))
    expect_true("param" %in% names(fmt))
    expect_true("value" %in% names(fmt))
})

#' Validate a data frame against a schema
#'
#' Checks that a data frame conforms to a schema, validating column presence,
#' types, missing values, uniqueness constraints, and nesting structure.
#'
#' @param schema A schema object created by [sch_schema()].
#' @param data A data frame to validate.
#' @param check A character vector specifying which checks to perform. The
#'   default runs all checks. Possible values:
#'   - `"names"`: check for missing required columns and unexpected extra columns.
#'   - `"types"`: check column types and missing-value (`NA`) constraints.
#'   - `"distinct"`: check uniqueness constraints for columns marked
#'     `distinct = TRUE`. Relatively expensive.
#'   - `"relationships"`: validate relationship formulas (primary-key uniqueness
#'     and crossing/nesting completeness). Relatively expensive.
#' @param call The environment or call used for error reporting, passed to
#'   [rlang::abort()]. Useful when wrapping `sch_validate()` inside another
#'   function so that the error points to the right place.
#'
#' @returns `data`, invisibly, if validation succeeds. Otherwise, an error of
#'   class `sch_validation_error` is raised with a formatted summary of all
#'   issues found.
#'
#' @examples
#' # Basic validation: valid data passes silently
#' schema <- sch_schema(
#'     id = sch_integer(distinct = TRUE),
#'     name = sch_character(missing = FALSE),
#'     age = sch_numeric(required = FALSE)
#' )
#' df <- data.frame(id = 1:3, name = c("Alice", "Bob", "Carol"), age = c(25, NA, 30))
#' sch_validate(schema, df)
#'
#' # Invalid data throws validation errors; wrap in try() so examples can run
#' # missing required columns
#' try(sch_validate(schema, data.frame(id = 1:2)))
#'
#' # type constraints not satisfied
#' try(sch_validate(schema, data.frame(id = c(1L, 1L), name = c("Alice", NA))))
#' @export
sch_validate <- function(
    schema,
    data,
    check = c("names", "types", "distinct", "relationships"),
    call = rlang::caller_env()
) {
    assert(
        inherits(schema, "sch_schema"),
        "{.arg schema} must be an {.cls sch_schema} object."
    )
    assert(is.data.frame(data), "{.arg data} must be a data frame.")

    check = rlang::arg_match(check, multiple = TRUE)
    col_info <- classify_columns(schema$cols)
    issues <- list()

    issues <- c(issues, validate_cols(schema$cols, data, col_info, check))

    if (any(c("names", "types", "distinct") %in% check)) {
        issues <- c(issues, validate_nests(schema$cols, data, col_info, check))
    }
    if ("relationships" %in% check) {
        issues <- c(issues, validate_relationships(schema, data))
    }

    if (length(issues) > 0) {
        classes = setdiff(class(data), c("sch_df", "tbl", "tbl_df", "data.frame", "data.table"))
        print_validation_issues(issues, class = classes, call = call)
    }

    invisible(data)
}


# Validators -------------------------------------------

classify_columns <- function(cols) {
    is_other <- vapply(cols, function(x) x$type == "other", FALSE)
    is_nest <- vapply(cols, function(x) x$type == "schema_nest", FALSE)
    is_multiple <- vapply(cols, function(x) x$type == "schema_multiple", FALSE)
    nms <- names(cols)
    if (is.null(nms)) {
        nms <- rep("", length(cols))
    }
    is_named_nest <- is_nest & nzchar(nms)

    list(
        nms = nms,
        is_other = is_other,
        is_nest = is_nest,
        is_multiple = is_multiple,
        is_named_nest = is_named_nest
    )
}


# Run the per-column validators (names, types, distinct) for a set of columns
validate_cols <- function(cols, data, col_info, check, path = character(0)) {
    issues <- list()
    if ("names" %in% check) {
        issues <- c(issues, validate_names(cols, data, col_info, path))
    }
    if ("types" %in% check) {
        issues <- c(issues, validate_types_missing(cols, data, col_info, path))
    }
    if ("distinct" %in% check) {
        issues <- c(issues, validate_distinct(cols, data, col_info, path))
    }
    issues
}


validate_names <- function(cols, data, col_info, path = character(0)) {
    issues <- list()
    groups_attr <- attr(data, "sch_groups")

    if (any(col_info$is_multiple) && is.null(groups_attr)) {
        issues <- c(issues, list(make_issue("missing_sch_groups", path)))
    }

    expected_names <- character(0)
    for (i in which(!col_info$is_other)) {
        if (col_info$is_multiple[i]) {
            if (!is.null(groups_attr)) {
                res <- check_multiple_names(cols[[i]], data, groups_attr, path)
                issues <- c(issues, res$issues)
                expected_names <- c(expected_names, res$expected_nms)
            }
        } else {
            col_nm <- col_info$nms[i]
            expected_names <- c(expected_names, col_nm)
            if (
                !col_info$is_named_nest[i] &&
                    isTRUE(attr(cols[[i]], "required")) &&
                    !col_nm %in% names(data)
            ) {
                new_issue = make_issue(
                    "missing_column",
                    c(path, col_nm),
                    expected = a_an_msg(cols[[i]])
                )
                issues <- c(issues, list(new_issue))
            }
        }
    }

    if (!any(col_info$is_other)) {
        extra <- setdiff(names(data), expected_names)
        if (length(extra) > 0) {
            issues <- c(issues, list(make_issue("extra_columns", path, columns = extra)))
        }
    }

    issues
}


check_multiple_names <- function(tt, data, groups_attr, path) {
    if (!tt$name %in% names(groups_attr)) {
        return(list(
            issues = list(make_issue("missing_group", path, name = tt$name)),
            expected_nms = character(0)
        ))
    }

    col_nms <- groups_attr[[tt$name]]
    issues <- list()
    if (length(col_nms) == 0 && isTRUE(attr(tt, "required"))) {
        issues <- c(issues, list(make_issue("empty_group", path, name = tt$name)))
    }
    for (col_nm in col_nms[!col_nms %in% names(data)]) {
        new_issue <- make_issue(
            "missing_column",
            c(path, col_nm),
            expected = a_an_msg(tt$inner)
        )
        issues <- c(issues, list(new_issue))
    }

    list(issues = issues, expected_nms = col_nms)
}


validate_types_missing <- function(cols, data, col_info, path = character(0)) {
    issues <- list()
    groups_attr <- attr(data, "sch_groups")

    for (i in which(!col_info$is_other & !col_info$is_nest)) {
        if (col_info$is_multiple[i]) {
            issues <- c(issues, check_multiple_types(cols[[i]], data, groups_attr, path))
        } else {
            col_nm <- col_info$nms[i]
            if (col_nm %in% names(data)) {
                issues <- c(
                    issues,
                    check_col_type_and_na(data[[col_nm]], cols[[i]], c(path, col_nm))
                )
            }
        }
    }

    issues
}


check_multiple_types <- function(tt, data, groups_attr, path) {
    if (is.null(groups_attr) || !tt$name %in% names(groups_attr)) {
        return(list())
    }
    col_nms <- groups_attr[[tt$name]]
    if (length(col_nms) == 0) {
        return(list())
    }

    issues <- list()
    per_col_ok <- TRUE
    for (col_nm in col_nms) {
        if (!col_nm %in% names(data)) {
            per_col_ok <- FALSE
            next
        }
        col_issues <- check_col_type_and_na(data[[col_nm]], tt$inner, c(path, col_nm))
        issues <- c(issues, col_issues)
        if (length(col_issues) > 0) per_col_ok <- FALSE
    }

    if (per_col_ok && !is.null(tt$cross_check)) {
        present_nms <- intersect(col_nms, names(data))
        col_list <- stats::setNames(lapply(present_nms, function(nm) data[[nm]]), present_nms)
        if (!isTRUE(tt$cross_check(col_list, tt))) {
            new_issue <- make_issue(
                "cross_check_failed",
                path,
                name = tt$name,
                expected = tt$cross_msg(tt)
            )
            issues <- c(issues, list(new_issue))
        }
    }

    issues
}


validate_distinct <- function(cols, data, col_info, path = character(0)) {
    issues <- list()

    for (i in seq_along(cols)) {
        if (col_info$is_other[i] || col_info$is_nest[i]) {
            next
        }

        if (col_info$is_multiple[i]) {
            tt <- cols[[i]]
            if (!isTRUE(attr(tt$inner, "distinct"))) {
                next
            }
            groups_attr <- attr(data, "sch_groups")
            if (is.null(groups_attr) || !tt$name %in% names(groups_attr)) {
                next
            }
            for (col_nm in groups_attr[[tt$name]]) {
                if (!col_nm %in% names(data)) {
                    next
                }
                issues <- c(issues, check_col_distinct(data[[col_nm]], c(path, col_nm)))
            }
            next
        }

        tt <- cols[[i]]
        if (!attr(tt, "distinct")) {
            next
        }
        col_nm <- col_info$nms[i]
        issues <- c(issues, check_col_distinct(data[[col_nm]], c(path, col_nm)))
    }

    issues
}


validate_nests <- function(cols, data, col_info, check, path = character(0)) {
    issues <- list()

    for (i in which(col_info$is_named_nest)) {
        issues <- c(
            issues,
            validate_named_nest(cols[[i]], col_info$nms[i], data, check, path)
        )
    }

    issues
}


validate_named_nest <- function(nest, col_nm, data, check, path) {
    issues <- list()
    col_path <- c(path, col_nm)

    if (!col_nm %in% names(data)) {
        if ("names" %in% check && attr(nest, "required")) {
            issues <- c(
                issues,
                list(make_issue("missing_column", col_path, expected = "nested data frame"))
            )
        }
        return(issues)
    }

    x <- data[[col_nm]]

    if (!is.list(x)) {
        issues <- c(issues, list(make_issue("not_nested_df", col_path)))
        return(issues)
    }

    # Validate each nested data frame element
    inner_col_info <- classify_columns(nest$cols)
    for (idx in seq_along(x)) {
        elem <- x[[idx]]
        if (!is.data.frame(elem)) {
            issues <- c(
                issues,
                list(make_issue("not_data_frame_element", col_path, index = idx))
            )
            next
        }

        inner_issues <- c(
            validate_cols(nest$cols, elem, inner_col_info, check, col_path),
            validate_nests(nest$cols, elem, inner_col_info, check, col_path)
        )

        for (j in seq_along(inner_issues)) {
            inner_issues[[j]]$element <- idx
        }
        issues <- c(issues, inner_issues)
    }

    issues
}


# Relationship validation -------------------------------------------------

validate_relationships <- function(schema, data) {
    tree <- schema$relationships
    if (is.null(tree)) {
        return(list())
    }

    # All formula columns
    all_cols <- relationship_columns(tree)

    # Skip if any formula column missing from data (names check catches it)
    if (!all(all_cols %in% names(data))) {
        return(list())
    }

    # Skip empty data frames
    if (nrow(data) == 0L) {
        return(list())
    }

    issues <- list()

    # Uniqueness: all formula columns form a unique key
    key_data <- data[, all_cols, drop = FALSE]
    grp <- vctrs::vec_group_loc(key_data)
    dup_mask <- lengths(grp$loc) > 1L
    if (any(dup_mask)) {
        n_dup <- nrow(data) - nrow(vctrs::vec_unique(key_data))
        first_key <- vctrs::vec_slice(grp$key, which(dup_mask)[1L])
        issues <- c(
            issues,
            list(make_issue(
                "duplicate_key",
                character(0),
                columns = all_cols,
                n_duplicates = n_dup,
                first_key = first_key
            ))
        )
    }

    issues <- c(issues, validate_rel_node(tree, data, group_cols = character(0)))

    issues
}


# Recursively validate a relationship tree node
validate_rel_node <- function(node, data, group_cols) {
    switch(
        node$type,
        var = list(),
        compound = list(),
        cross = validate_rel_cross(node, data, group_cols),
        nest = validate_rel_nest(node, data, group_cols),
        rlang::abort(paste0("Unknown relationship node type: ", node$type))
    )
}


# Check crossing completeness: within each group defined by group_cols,
# the unique combos of all children = product of unique combos per child.
# When group_cols is non-empty, all failing groups are summarized into a
# single issue showing the count and the first offending group key.
validate_rel_cross <- function(node, data, group_cols) {
    issues <- list()
    children <- node$children

    # Get column names per child
    child_cols <- lapply(children, relationship_columns)
    label <- paste(
        vapply(child_cols, function(cc) paste(cc, collapse = " + "), ""),
        collapse = " \u00d7 "
    )

    if (length(group_cols) == 0L) {
        issues <- c(issues, check_cross_completeness(data, child_cols, label, NULL, 1L, 1L))
    } else {
        grp_data <- data[, group_cols, drop = FALSE]
        locs <- vctrs::vec_group_loc(grp_data)
        n_groups <- length(locs$loc)
        n_fail <- 0L
        first_key <- NULL
        first_actual <- NULL
        first_expected <- NULL

        for (i in seq_len(n_groups)) {
            sub <- vctrs::vec_slice(data, locs$loc[[i]])
            result <- cross_completeness_counts(sub, child_cols)
            if (result$actual < result$expected) {
                n_fail <- n_fail + 1L
                if (is.null(first_key)) {
                    first_key <- vctrs::vec_slice(grp_data, locs$loc[[i]][1L])
                    first_actual <- result$actual
                    first_expected <- result$expected
                }
            }
        }

        if (n_fail > 0L) {
            issues <- c(
                issues,
                list(make_issue(
                    "incomplete_crossing",
                    character(0),
                    label = label,
                    actual = first_actual,
                    expected = first_expected,
                    group_cols = group_cols,
                    first_key = first_key,
                    n_fail = n_fail,
                    n_groups = n_groups
                ))
            )
        }
    }

    # Recurse into non-leaf children only; cross completeness handles var/compound
    for (k in seq_along(children)) {
        child <- children[[k]]
        if (child$type %in% c("var", "compound")) {
            next
        }
        sibling_cols <- unique(unlist(child_cols[-k]))
        new_group <- c(group_cols, sibling_cols)
        issues <- c(issues, validate_rel_node(child, data, new_group))
    }

    issues
}


# Returns list(actual, expected) counts for crossing completeness
cross_completeness_counts <- function(data, child_cols) {
    all_cols <- unique(unlist(child_cols))
    actual <- nrow(vctrs::vec_unique(data[, all_cols, drop = FALSE]))
    expected <- prod(vapply(
        child_cols,
        function(cc) nrow(vctrs::vec_unique(data[, cc, drop = FALSE])),
        0L
    ))
    list(actual = actual, expected = expected)
}


# Check that unique(all child cols) == product of unique(each child's cols)
check_cross_completeness <- function(data, child_cols, label, first_key, n_fail, n_groups) {
    if (length(child_cols) < 2) {
        return(list())
    }

    result <- cross_completeness_counts(data, child_cols)

    if (result$actual < result$expected) {
        return(list(make_issue(
            "incomplete_crossing",
            character(0),
            label = label,
            actual = result$actual,
            expected = result$expected,
            group_cols = character(0),
            first_key = first_key,
            n_fail = n_fail,
            n_groups = n_groups
        )))
    }

    list()
}


# Check nesting: inner is scoped within outer groups
validate_rel_nest <- function(node, data, group_cols) {
    issues <- list()
    outer_cols <- relationship_columns(node$outer)

    # For three-level (or deeper) formulas where outer is itself a nest and
    # inner is a cross, sibling groups at the intermediate level must carry
    # consistent inner structure. The scope for this consistency check
    # depends on the structure of the intermediate cross:
    #
    #   Pure cross (no nested children): groups are fully symmetric, so scope
    #   to the grandparent (outer$outer) level. All siblings within a
    #   grandparent group must share the same inner grid.
    #
    #   Mixed cross (some nested children): the non-nested children define a
    #   "symmetric" dimension and the nested children define a heterogeneous
    #   one. Extend the grandparent scope by the non-nested children's columns
    #   so that only groups sharing both the grandparent identity AND the
    #   symmetric dimension are compared. E.g. for
    #   `(race) / (geo * (party / candidate)) / (time * method)`, scope is
    #   per (race, geo): within a geo, all (party, candidate) groups must have
    #   the same time × method grid, but different geos may have different grids.
    #
    # All failures across every scope group are aggregated into a single issue
    # so the user sees one summary rather than one error per scope group.
    if (node$outer$type == "nest" && node$inner$type == "cross") {
        first_outer_cols <- relationship_columns(node$outer$outer)
        scope_extension <- character(0L)

        if (node$outer$inner$type == "cross") {
            outer_inner_children <- node$outer$inner$children
            is_nest_ch <- vapply(outer_inner_children, function(ch) ch$type == "nest", logical(1L))
            if (any(is_nest_ch)) {
                scope_extension <- unlist(lapply(
                    outer_inner_children[!is_nest_ch],
                    relationship_columns
                ))
            }
        }

        scope_cols <- unique(c(group_cols, first_outer_cols, scope_extension))

        # Column info for the inner cross node (for reference computation & label)
        child_cols <- lapply(node$inner$children, relationship_columns)
        all_child_cols <- unique(unlist(child_cols))
        label <- paste(
            vapply(child_cols, function(cc) paste(cc, collapse = " + "), ""),
            collapse = " \u00d7 "
        )

        # Aggregate failures across all scope groups into a single issue
        n_fail <- 0L
        n_groups <- 0L
        first_key <- NULL
        first_actual <- NULL
        first_expected <- NULL

        run_scope <- function(sub_scope) {
            ref_expected <- cross_completeness_counts(sub_scope, child_cols)$expected
            sub_grp <- sub_scope[, outer_cols, drop = FALSE]
            sub_locs <- vctrs::vec_group_loc(sub_grp)
            n_groups <<- n_groups + length(sub_locs$loc)
            for (i in seq_len(length(sub_locs$loc))) {
                sub <- vctrs::vec_slice(sub_scope, sub_locs$loc[[i]])
                actual <- nrow(vctrs::vec_unique(sub[, all_child_cols, drop = FALSE]))
                if (actual < ref_expected) {
                    n_fail <<- n_fail + 1L
                    if (is.null(first_key)) {
                        first_key <<- vctrs::vec_slice(sub_grp, sub_locs$loc[[i]][1L])
                        first_actual <<- actual
                        first_expected <<- ref_expected
                    }
                }
            }
        }

        if (length(scope_cols) == 0L) {
            run_scope(data)
        } else {
            scope_data <- data[, scope_cols, drop = FALSE]
            scope_locs <- vctrs::vec_group_loc(scope_data)
            for (si in seq_len(length(scope_locs$loc))) {
                run_scope(vctrs::vec_slice(data, scope_locs$loc[[si]]))
            }
        }

        if (n_fail > 0L) {
            issues <- c(
                issues,
                list(make_issue(
                    "incomplete_crossing",
                    character(0),
                    label = label,
                    actual = first_actual,
                    expected = first_expected,
                    group_cols = outer_cols,
                    first_key = first_key,
                    n_fail = n_fail,
                    n_groups = n_groups
                ))
            )
        }

        # Recurse into non-leaf children of the inner cross
        for (k in seq_along(node$inner$children)) {
            child <- node$inner$children[[k]]
            if (child$type %in% c("var", "compound")) {
                next
            }
            sibling_cols <- unique(unlist(child_cols[-k]))
            issues <- c(issues, validate_rel_node(child, data, c(group_cols, sibling_cols)))
        }
    } else {
        issues <- c(issues, validate_rel_node(node$inner, data, outer_cols))
    }

    # Only recurse into outer if it has sub-structure (cross or nest);
    # var/compound outers are pure grouping context, not independently validated.
    if (node$outer$type %in% c("cross", "nest")) {
        issues <- c(issues, validate_rel_node(node$outer, data, group_cols))
    }

    issues
}


# Printing ---------------------------------------------------------------

print_validation_issues <- function(issues, class = NULL, call = NULL) {
    n <- length(issues)

    bullets <- vapply(issues, print_validation_issue, character(1))
    names(bullets) <- rep("x", n)
    if (is.null(class) || length(class) == 0) {
        class <- "Data"
    } else {
        class = paste0("{.cls ", paste(class, collapse = "}/{.cls "), "}")
    }

    cli::cli_abort(
        c(paste(class, "validation failed with {n} issue{?s}:"), bullets),
        class = "sch_validation_error",
        issues = issues,
        call = call
    )
}

print_validation_issue <- function(issue) {
    p <- paste(issue$path, collapse = "$")
    elem_note <- if (!is.null(issue$element)) {
        paste0(" (element ", issue$element, ")")
    } else {
        ""
    }

    switch(
        issue$type,
        missing_column = cli::format_inline(
            "Required column {.field {p}}{elem_note} is missing. Expected {issue$expected}."
        ),
        wrong_type = cli::format_inline(
            "Column {.field {p}}{elem_note} has wrong type. Expected {issue$expected}."
        ),
        has_na = cli::format_inline(
            "Column {.field {p}}{elem_note} must not contain missing values."
        ),
        not_distinct = cli::format_inline(
            "Column {.field {p}}{elem_note} must not contain duplicate values."
        ),
        extra_columns = {
            nms <- cli::cli_vec(issue$columns, list("vec-trunc" = 5))
            cli::format_inline("Unexpected column{?s}: {.field {nms}}.")
        },
        not_nested_df = cli::format_inline(
            "Column {.field {p}} should be a list of data frames."
        ),
        not_data_frame_element = cli::format_inline(
            "Column {.field {p}} element {issue$index} is not a data frame."
        ),
        missing_sch_groups = cli::format_inline(
            "Data frame is missing the {.field sch_groups} attribute (required by {.fn sch_multiple})."
        ),
        missing_group = cli::format_inline(
            "Malformed attributes: group {.field {issue$name}} not found in {.field sch_groups}."
        ),
        empty_group = cli::format_inline(
            "Malformed attributes: group {.field {issue$name}} in {.field sch_groups} is emptpy but {.code required = TRUE})."
        ),
        cross_check_failed = cli::format_inline(
            "Column group {.field {issue$name}} failed cross-column check. Expected {issue$expected}."
        ),
        duplicate_key = {
            cols_str <- paste(issue$columns, collapse = ", ")
            key_str <- format_key(issue$first_key)
            cli::format_inline(
                "Columns ({.field {cols_str}}) should be unique per row. Found {issue$n_duplicates} duplicate combination{?s}. First duplicate: ({key_str})."
            )
        },
        incomplete_crossing = {
            if (length(issue$group_cols) == 0L) {
                cli::format_inline(
                    "Incomplete crossing of {issue$label}: found {issue$actual} of {issue$expected} expected combinations."
                )
            } else {
                grp_label <- paste(issue$group_cols, collapse = " + ")
                key_str <- format_key(issue$first_key)
                if (issue$n_fail == 1L) {
                    cli::format_inline(
                        "Incomplete crossing of {issue$label} in ({.field {grp_label}}) group ({key_str}): found {issue$actual} of {issue$expected} expected combinations."
                    )
                } else {
                    cli::format_inline(
                        "Incomplete crossing of {issue$label} in {issue$n_fail} of {issue$n_groups} ({.field {grp_label}}) groups. First at ({key_str}): found {issue$actual} of {issue$expected} expected combinations."
                    )
                }
            }
        },
        paste0("Unknown issue: ", issue$type)
    )
}


# Helpers ----------

make_issue <- function(type, path, ...) {
    c(list(type = type, path = path), list(...))
}

# Format a single-row data frame key as "col1=val1, col2=val2, ..."
format_key <- function(key) {
    if (is.null(key) || ncol(key) == 0L) {
        return("")
    }
    parts <- mapply(
        function(nm, val) paste0(nm, "=", as.character(val[[1L]])),
        names(key),
        as.list(key),
        SIMPLIFY = TRUE
    )
    paste(parts, collapse = ", ")
}

# Build "a <type description>" / "an <type description>" for a type object
a_an_msg <- function(tt) {
    msg <- type_fns[[tt$type]]$msg(tt)
    paste0(if (grepl("^[aeiou]", msg)) "an " else "a ", msg)
}

# TRUE if a vector or list-column contains any NA / NULL elements
has_missing <- function(x) {
    if (is.list(x)) anyNA(x) || any(vapply(x, is.null, logical(1))) else anyNA(x)
}

# Type check + (gated) NA check for a single column vector
check_col_type_and_na <- function(x, tt, col_path) {
    issues <- list()
    type_ok <- isTRUE(type_fns[[tt$type]]$check(x, tt))
    if (!type_ok) {
        issues <- c(issues, list(make_issue("wrong_type", col_path, expected = a_an_msg(tt))))
    }
    if (type_ok && !attr(tt, "missing") && has_missing(x)) {
        issues <- c(issues, list(make_issue("has_na", col_path)))
    }
    issues
}

# Distinctness check for a single column vector
check_col_distinct <- function(x, col_path) {
    x_obs <- x[!is.na(x)]
    if (vctrs::vec_unique_count(x_obs) != vctrs::vec_size(x_obs)) {
        list(make_issue("not_distinct", col_path))
    } else {
        list()
    }
}

#' Coerce a data frame to conform to a schema
#'
#' Attempts to coerce each column of `data` to the type expected by `schema`,
#' using the coercion method defined for each column type. After coercion,
#' optionally validates the result with [sch_validate()].
#'
#' Coercion is applied column-by-column using the `coerce` function registered
#' for each type in the internal `type_fns` registry. For example, a column
#' specified as `sch_integer()` will be coerced with [as.integer()]. Nested
#' schemas (created with [sch_nest()]) are handled by recursing into each
#' element data frame, and grouped columns (created with [sch_multiple()]) have
#' each member column coerced individually.
#'
#' Columns present in `data` but not named in `schema` (i.e., those covered by
#' [sch_others()]) are left untouched.
#'
#' @param schema A schema object created by [sch_schema()].
#' @param data A data frame to coerce.
#' @param validate If `TRUE` (default), [sch_validate()] is called on the
#'   coerced data after all columns have been processed. Set to `FALSE` to skip
#'   validation, which can be useful when you want to inspect the coerced result
#'   before checking constraints.
#' @param call The environment or call used for error reporting, passed to
#'   [rlang::abort()]. Useful when wrapping `sch_coerce()` inside another
#'   function so that errors point to the right place.
#'
#' @returns `data` with columns coerced to their schema types, invisibly, if
#'   coercion succeeds. If any column cannot be coerced, an error of class
#'   `sch_coercion_error` is raised with a summary of all failures. When
#'   `validate = TRUE`, a subsequent [sch_validate()] call may also raise a
#'   `sch_validation_error` if the coerced data still violates schema
#'   constraints (e.g., out-of-bounds values or uniqueness violations).
#'
#' @examples
#' schema <- sch_schema(
#'     id = sch_integer(distinct = TRUE),
#'     name = sch_character(missing = FALSE),
#'     score = sch_numeric()
#' )
#'
#' # Coerce a data frame with character columns
#' df <- data.frame(id = c("1", "2", "3"), name = c("Alice", "Bob", "Carol"), score = 1:3)
#' str(sch_coerce(schema, df))
#'
#' # Nested schema coercion
#' nested_schema <- sch_schema(
#'     group = sch_factor(),
#'     info = sch_nest(x = sch_numeric(), y = sch_integer())
#' )
#' nested_df <- data.frame(group = "A")
#' nested_df$info <- list(data.frame(x = "1.5", y = "2"))
#' str(sch_coerce(nested_schema, nested_df))
#' @export
sch_coerce <- function(schema, data, validate = TRUE, call = rlang::caller_env()) {
    assert(
        inherits(schema, "sch_schema"),
        "{.arg schema} must be an {.cls sch_schema} object."
    )
    assert(is.data.frame(data), "{.arg data} must be a data frame.")

    result <- coerce_cols(schema$cols, data, path = character(0))
    data <- result$data
    issues <- result$issues

    if (length(issues) > 0) {
        print_coercion_issues(issues, call = call)
    }

    if (isTRUE(validate)) {
        sch_validate(schema, data, call = call)
    }

    data
}


# Internal helpers --------------------------------------------------------

# Coerce all schema columns in `data`, returning list(data, issues).
coerce_cols <- function(cols, data, path = character(0)) {
    col_info <- classify_columns(cols)
    issues <- list()

    for (i in seq_along(cols)) {
        tt <- cols[[i]]

        if (col_info$is_other[i]) {
            next
        }

        if (col_info$is_named_nest[i]) {
            col_nm <- col_info$nms[i]
            if (col_nm %in% names(data)) {
                result <- coerce_nest_col(tt, data[[col_nm]], col_nm, path)
                data[[col_nm]] <- result$value
                issues <- c(issues, result$issues)
            }
            next
        }

        if (col_info$is_multiple[i]) {
            groups_attr <- attr(data, "sch_groups")
            if (!is.null(groups_attr) && tt$name %in% names(groups_attr)) {
                for (col_nm in groups_attr[[tt$name]]) {
                    if (col_nm %in% names(data)) {
                        result <- coerce_one(data[[col_nm]], tt$inner, c(path, col_nm))
                        data[[col_nm]] <- result$value
                        issues <- c(issues, result$issues)
                    }
                }
            }
            next
        }

        col_nm <- col_info$nms[i]
        if (col_nm %in% names(data)) {
            result <- coerce_one(data[[col_nm]], tt, c(path, col_nm))
            data[[col_nm]] <- result$value
            issues <- c(issues, result$issues)
        }
    }

    list(data = data, issues = issues)
}


# Coerce a single column vector; return list(value, issues).
coerce_one <- function(x, tt, col_path) {
    coerce_fn <- type_fns[[tt$type]]$coerce
    tryCatch(
        list(value = coerce_fn(x, tt), issues = list()),
        error = function(e) {
            list(
                value = x,
                issues = list(make_coercion_issue(col_path, conditionMessage(e)))
            )
        }
    )
}


# Recursively coerce a list-column of data frames according to a nest spec.
coerce_nest_col <- function(nest, list_col, col_nm, path) {
    issues <- list()
    col_path <- c(path, col_nm)

    for (idx in seq_along(list_col)) {
        elem <- list_col[[idx]]
        if (!is.data.frame(elem)) {
            next
        }
        result <- coerce_cols(nest$cols, elem, path = col_path)
        list_col[[idx]] <- result$data
        elem_issues <- result$issues
        for (j in seq_along(elem_issues)) {
            elem_issues[[j]]$element <- idx
        }
        issues <- c(issues, elem_issues)
    }

    list(value = list_col, issues = issues)
}


make_coercion_issue <- function(path, message) {
    list(path = path, message = message)
}


print_coercion_issues <- function(issues, call = NULL) {
    n <- length(issues)
    bullets <- vapply(
        issues,
        function(issue) {
            p <- paste(issue$path, collapse = "$")
            elem_note <- if (!is.null(issue$element)) {
                paste0(" (element ", issue$element, ")")
            } else {
                ""
            }
            cli::format_inline("Column {.field {p}}{elem_note}: {issue$message}")
        },
        character(1)
    )
    names(bullets) <- rep("x", n)
    cli::cli_abort(
        c("Coercion failed for {n} column{?s}:", bullets),
        class = "sch_coercion_error",
        issues = issues,
        call = call
    )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{sch_coerce}
\alias{sch_coerce}
\title{Coerce a data frame to conform to a schema}
\usage{
sch_coerce(schema, data, validate = TRUE, call = rlang::caller_env())
}
\arguments{
\item{schema}{A schema object created by [sch_schema()].}

\item{data}{A data frame to coerce.}

\item{validate}{If `TRUE` (default), [sch_validate()] is called on the
coerced data after all columns have been processed. Set to `FALSE` to skip
validation, which can be useful when you want to inspect the coerced result
before checking constraints.}

\item{call}{The environment or call used for error reporting, passed to
[rlang::abort()]. Useful when wrapping `sch_coerce()` inside another
function so that errors point to the right place.}
}
\value{
`data` with columns coerced to their schema types, invisibly, if
  coercion succeeds. If any column cannot be coerced, an error of class
  `sch_coercion_error` is raised with a summary of all failures. When
  `validate = TRUE`, a subsequent [sch_validate()] call may also raise a
  `sch_validation_error` if the coerced data still violates schema
  constraints (e.g., out-of-bounds values or uniqueness violations).
}
\description{
Attempts to coerce each column of `data` to the type expected by `schema`,
using the coercion method defined for each column type. After coercion,
optionally validates the result with [sch_validate()].
}
\details{
Coercion is applied column-by-column using the `coerce` function registered
for each type in the internal `type_fns` registry. For example, a column
specified as `sch_integer()` will be coerced with [as.integer()]. Nested
schemas (created with [sch_nest()]) are handled by recursing into each
element data frame, and grouped columns (created with [sch_multiple()]) have
each member column coerced individually.

Columns present in `data` but not named in `schema` (i.e., those covered by
[sch_others()]) are left untouched.
}
\examples{
schema <- sch_schema(
    id = sch_integer(distinct = TRUE),
    name = sch_character(missing = FALSE),
    score = sch_numeric()
)

# Coerce a data frame with character columns
df <- data.frame(id = c("1", "2", "3"), name = c("Alice", "Bob", "Carol"), score = 1:3)
str(sch_coerce(schema, df))

# Nested schema coercion
nested_schema <- sch_schema(
    group = sch_factor(),
    info = sch_nest(x = sch_numeric(), y = sch_integer())
)
nested_df <- data.frame(group = "A")
nested_df$info <- list(data.frame(x = "1.5", y = "2"))
str(sch_coerce(nested_schema, nested_df))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr.R
\name{dplyr-sch_df}
\alias{dplyr-sch_df}
\title{dplyr integration for sch_df}
\description{
These methods hook into dplyr's three extension generics
([dplyr::dplyr_row_slice()], [dplyr::dplyr_col_modify()],
[dplyr::dplyr_reconstruct()]) plus the base-R [names()] replacement and
1-d `[` to enforce schema constraints across dplyr operations with near-zero
overhead.
}
\details{
Each method calls `NextMethod()` first (performing the actual dplyr
operation) and then re-validates the result against the schema, running only
the checks that can plausibly be violated by that type of operation:

| Operation | Checks run |
|---|---|
| `dplyr_row_slice` | none by default; `"relationships"` behind a flag |
| `dplyr_col_modify` | `"names"`, `"types"`, `"distinct"` |
| `dplyr_reconstruct` | `"names"`, `"types"` |
| `[` (1-d column subsetting) | `"names"` |
| `names<-` | errors if any *schema* column name is changed |
}
\section{Row slicing (`arrange`, `filter`, `slice`, semi/anti joins)}{

Row operations cannot introduce new name or type violations, so no
validation is run by default. Relationship constraints (crossing
completeness, primary-key uniqueness) *can* be broken by removing rows.
Pass `.check_relationships = TRUE` in `...` to opt in to that check.
}

\section{Column modification (`mutate`)}{

A column modification can delete required columns (via `NULL` assignment),
assign the wrong type, or introduce duplicate values into a `distinct = TRUE`
column, so all three cheap checks are run.
}

\section{Reconstruction (joins)}{

`dplyr_reconstruct()` is called after joins. Only names and types are
checked: distinct and relationship checks are omitted because joins can
intentionally produce non-distinct rows or incomplete crossings.
}

\section{Column subsetting (`select`, `relocate`)}{

A 1-d `[` call selects or reorders columns. Reordering is always safe; but
selecting a column subset could drop required columns, so a names check is
run.
}

\section{Renaming (`rename`, `rename_with`, `select` with rename)}{

Renaming a column that belongs to the schema (either directly or as a member
of a [sch_multiple()] group) is never valid without also updating the schema
definition, so an error is raised immediately.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate.R
\name{sch_validate}
\alias{sch_validate}
\title{Validate a data frame against a schema}
\usage{
sch_validate(
  schema,
  data,
  check = c("names", "types", "distinct", "relationships"),
  call = rlang::caller_env()
)
}
\arguments{
\item{schema}{A schema object created by [sch_schema()].}

\item{data}{A data frame to validate.}

\item{check}{A character vector specifying which checks to perform. The
default runs all checks. Possible values:
- `"names"`: check for missing required columns and unexpected extra columns.
- `"types"`: check column types and missing-value (`NA`) constraints.
- `"distinct"`: check uniqueness constraints for columns marked
  `distinct = TRUE`. Relatively expensive.
- `"relationships"`: validate relationship formulas (primary-key uniqueness
  and crossing/nesting completeness). Relatively expensive.}

\item{call}{The environment or call used for error reporting, passed to
[rlang::abort()]. Useful when wrapping `sch_validate()` inside another
function so that the error points to the right place.}
}
\value{
`data`, invisibly, if validation succeeds. Otherwise, an error of
  class `sch_validation_error` is raised with a formatted summary of all
  issues found.
}
\description{
Checks that a data frame conforms to a schema, validating column presence,
types, missing values, uniqueness constraints, and nesting structure.
}
\examples{
# Basic validation: valid data passes silently
schema <- sch_schema(
    id = sch_integer(distinct = TRUE),
    name = sch_character(missing = FALSE),
    age = sch_numeric(required = FALSE)
)
df <- data.frame(id = 1:3, name = c("Alice", "Bob", "Carol"), age = c(25, NA, 30))
sch_validate(schema, df)

# Invalid data throws validation errors; wrap in try() so examples can run
# missing required columns
try(sch_validate(schema, data.frame(id = 1:2)))

# type constraints not satisfied
try(sch_validate(schema, data.frame(id = c(1L, 1L), name = c("Alice", NA))))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schema.R
\name{sch_schema}
\alias{sch_schema}
\alias{sch_others}
\alias{sch_any}
\alias{sch_multiple}
\alias{sch_nest}
\alias{sch_numeric}
\alias{sch_integer}
\alias{sch_logical}
\alias{sch_character}
\alias{sch_factor}
\alias{sch_date}
\alias{sch_datetime}
\alias{sch_inherits}
\alias{sch_list_of}
\alias{sch_custom}
\title{Define a structured data type}
\usage{
sch_schema(..., .desc = NULL, .relationships = NULL)

sch_others()

sch_any(desc = NULL, missing = TRUE, required = TRUE, distinct = FALSE)

sch_multiple(
  name,
  type,
  desc = NULL,
  required = TRUE,
  check = NULL,
  msg = NULL,
  coerce = NULL
)

sch_nest(..., .desc = NULL)

sch_numeric(
  desc = NULL,
  bounds = c(-Inf, Inf),
  closed = c(TRUE, TRUE),
  missing = TRUE,
  required = TRUE,
  distinct = FALSE
)

sch_integer(
  desc = NULL,
  bounds = c(-Inf, Inf),
  closed = c(TRUE, TRUE),
  missing = TRUE,
  required = TRUE,
  distinct = FALSE
)

sch_logical(desc = NULL, missing = TRUE, required = TRUE, distinct = FALSE)

sch_character(desc = NULL, missing = TRUE, required = TRUE, distinct = FALSE)

sch_factor(
  desc = NULL,
  levels = NULL,
  strict = TRUE,
  missing = TRUE,
  required = TRUE,
  distinct = FALSE
)

sch_date(
  desc = NULL,
  bounds = c(as.Date(-Inf), as.Date(Inf)),
  closed = c(FALSE, FALSE),
  missing = TRUE,
  required = TRUE,
  distinct = FALSE
)

sch_datetime(
  desc = NULL,
  bounds = c(as.POSIXct(-Inf), as.POSIXct(Inf)),
  closed = c(FALSE, FALSE),
  missing = TRUE,
  required = TRUE,
  distinct = FALSE
)

sch_inherits(
  desc = NULL,
  class,
  missing = TRUE,
  required = TRUE,
  distinct = FALSE
)

sch_list_of(
  desc = NULL,
  class,
  missing = TRUE,
  required = TRUE,
  distinct = FALSE
)

sch_custom(
  name,
  desc = NULL,
  check,
  msg,
  coerce,
  ...,
  missing = TRUE,
  required = TRUE,
  distinct = FALSE
)
}
\arguments{
\item{...}{Column specifications, in the form of `col_name = col_type` pairs,
  where `col_type` is a call to a column type constructor listed here, such
  as `sch_numeric()`. Every type must be a kind of vector, i.e.,
  [vctrs::obj_is_vector()] must return `TRUE`.

  All columns must be named, except for `sch_others()`, as described below,
  and `sch_multiple()`, which describes a group of columns sharing the same
  type. A named `sch_nest()` describes columns stored as a nested data frame.

  The special function `sch_others()` indicates the preferred location of
  other columns not explicitly mentioned in the schema. If no `sch_others()`
  appears, then other columns are not allowed.
  Trailing commas are permitted.}

\item{.relationships}{An optional one-sided formula describing the structural
relationships between values in different columns. Formulas can only involve
named arguments to `...`. Use `*` to signify crossed levels, which will
verify all combinations exist, `/` to signify nested levels, and `+` to create
compound keys (bundling columns into a single identifier).
See the examples below.}

\item{desc, .desc}{A description of the column for consumers of the schema.
The type contraints will be described separately and do not need to be
included in the description.  For example for "age", the description might
be "Age of the patient in years", not "Non-negative integer representing
the age of the patient in years". For the overall `sch_schema`, the `desc`
will be printed as part of the header for data frames implementing the
schema, by default.}

\item{missing}{If `TRUE`, the column may be contain missing values. Otherwise,
any missing values result in an error.}

\item{required}{If `TRUE` (default), the group entry in `sch_groups` must
contain at least one column name. If `FALSE`, an empty character vector
for that entry is also accepted.}

\item{distinct}{If `TRUE`, the column must contain no duplicate values (after
accounting for nesting structure).}

\item{name}{A name for the custom type.}

\item{type}{A column type constructor (e.g. [sch_numeric()]) specifying the
expected type of every column in the group.}

\item{check}{A two-argument function that checks whether an object satisfies
the type. The first argument is the object to check, and the second is the
full type specification.}

\item{msg}{A one-argument function that generates a descriptive message
about the type when passed the type object itself. Should not end with a period.}

\item{coerce}{A two-argument function that attempts to coerce an object to the
type. The first argument is the object to coerce, and the second is the full
type specification.}

\item{bounds}{Length-two vector `c(min, max)` specifying the allowed range of values.}

\item{closed}{Length-two logical vector specifying whether the bounds are
closed (inclusive) or open (exclusive).}

\item{levels}{A character vector of factor levels, or NULL not enforce specific levels.}

\item{strict}{If `TRUE`, only factors with the specified levels are accepted.
If `FALSE`, character vectors with the specified levels are also accepted.}

\item{class}{A character vector of class names.}
}
\value{
An object of class `sch_schema`,
}
\description{
Defines the structure of a single 'observation' for a structured data frame.
Each column has type restrictions and may be required or optional.
Schemas support nesting relationships.
}
\section{Functions}{
\itemize{
\item \code{sch_others()}: A placeholder for other non-required columns in a schema.

\item \code{sch_any()}: A column of any type. No type checking is performed.

\item \code{sch_multiple()}: A group of multiple columns sharing the same type. The
  group is identified by `name`, which must appear as an entry in the
  `sch_groups` attribute of the data frame being validated. That entry is a
  character vector of column names that belong to this group.

  Optionally accepts cross-column `check`, `msg`, and `coerce` functions that
  are applied to the entire group after per-column type checks pass. These
  must all be provided together or not at all.

  `sch_multiple()` must be unnamed in an `sch_schema()` call. Per-column
  constraints such as `missing` and `distinct` are set on the inner `type`
  argument.

\item \code{sch_nest()}: A set of columns stored as a nested list-column of
data frames. Must be given a name in the outer `sch_schema()`.

\item \code{sch_numeric()}: A numeric vector that is optionally constrained to be
within a certain range.

\item \code{sch_integer()}: An integer vector that is optionally constrained to be
within a certain range.

\item \code{sch_logical()}: A logical vector.

\item \code{sch_character()}: A character vector.

\item \code{sch_factor()}: A factor with specified levels.

\item \code{sch_date()}: A Date vector that is optionally constrained to be
within a certain range.

\item \code{sch_datetime()}: A POSIXct vector that is optionally constrained to be
within a certain range.

\item \code{sch_inherits()}: A list-column whose elements satisfy `inherits(_, class)`.

\item \code{sch_list_of()}: A vector satisfying `inherits(_, class)`.

\item \code{sch_custom()}: A custom type defined by user-provided check, type message,
and coercion functions. Additional named values to be stored along with the
type specification may be passed via `...` and will be available to the
check, message, and coercion function as elements of the `type` argument.

}}
\section{Worked examples}{

Three larger end-to-end example scripts are bundled with the package and
show how to build a schema, attach group metadata, validate a compliant
data frame, and exercise a range of corruption cases. List them with:

```r
list.files(system.file("examples", package = "schmear"), full.names = TRUE)
```

The included scripts are `mcmc_draws.R` (fully-crossed MCMC posterior
draws), `ei_spec.R` (an ecological-inference specification using
[sch_multiple()] groups with a row-sum cross-column check), and
`election_data.R` (tidy election returns with a compound primary key and
nested/crossed relationship structure).
}

\examples{
sch_schema(
    .desc = "MCMC draws",
    .relationships = ~ chain * draw * parameter,
    chain = sch_integer("Chain number"),
    draw = sch_integer("Draw number", bounds = c(1, Inf), closed = c(TRUE, FALSE)),
    parameter = sch_factor("Parameter name", levels = c("mu", "sigma", "log_lik")),
    value = sch_numeric("Parameter value")
)

sch_schema(
    .desc = "Student data",
    .relationships = ~ (grade + teacher) / table_group,
    birthday = sch_date("Date of birth", required = FALSE),
    height = sch_numeric(
        "Height in inches",
        bounds = c(0, 108),
        closed = c(FALSE, TRUE)
    ),
    grade = sch_factor(strict = FALSE, levels = c("Kindergarten", "1st", "2nd")),
    teacher = sch_nest(
        first = sch_character("First name"),
        last = sch_character("Last name")
    ),
    table_group = sch_integer(bounds=c(1, 6)),
    enrolled = sch_logical(missing = FALSE),
    sch_others()
)

sch_schema(
    .desc = "Causal inference data",
    treatment = sch_factor(levels = c("control", "treatment"), missing = FALSE),
    outcome = sch_numeric(missing = FALSE),
    sch_multiple("covariates", type = sch_any(missing = FALSE), required = FALSE),
    sch_others()
)

sch_custom(
   name = "even",
   check = function(x, type) is.integer(x) && all(x \%\% 2 == 0),
   msg = function(type) "vector of even integers",
   coerce = function(x, type) (as.integer(x) \%/\% 2) * 2
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schmear-package.R
\docType{package}
\name{schmear-package}
\alias{schmear}
\alias{schmear-package}
\title{schmear: Build Structured Data Frame Subtypes}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Provides developer-focused helper functions and S3 classes to ease the creation of structured subtypes of data frames. Developers can require certain columns and types to be present, and can enforce crossing and nesting relationships between values in different columns. Type-specific metadata and attributes are preserved through common data frame manipulations.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://corymccartan.com/schmear/}
  \item \url{https://github.com/CoryMcCartan/schmear}
  \item Report bugs at \url{https://github.com/CoryMcCartan/schmear/issues}
}

}
\author{
\strong{Maintainer}: Cory McCartan \email{mccartan@psu.edu} (\href{https://orcid.org/0000-0002-6251-669X}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Cory McCartan \email{mccartan@psu.edu} (\href{https://orcid.org/0000-0002-6251-669X}{ORCID}) [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/df.R
\name{sch_df}
\alias{sch_df}
\alias{new_sch_df}
\alias{validate_sch_df}
\title{Construct and validate a schema-aware data frame}
\usage{
new_sch_df(data, schema, groups = NULL, class = NULL, use_tbl = TRUE)

validate_sch_df(x)
}
\arguments{
\item{data}{A data frame.}

\item{schema}{A `sch_schema` object.}

\item{groups}{A named list of character vectors of column names, for use
with [sch_multiple()].}

\item{class}{Additional classes to add to the object, in addition to "sch_df" and tibble/data.frame classes.}

\item{use_tbl}{If `TRUE`, the returned object will have tibble
classes `tbl_df` and `tbl` in addition to `data.frame`. As a reminder,
`tbl_df` affects the behavior of the object (slicing, row names, etc.),
while `tbl` affects printing only.}

\item{x}{A `sch_df` object.}
}
\value{
A data frame with class `c(class, "sch_df", ...)` and attributes
  `sch_schema` and `sch_groups`.

The input, if validation is successful.
}
\description{
Bare-bones constructor for a data frame with an attached schema. This function
should be called by package developers writing their own internal constructors.
The only checks are for the types of `data` and `schema`.
The `validate_sch_df()` is a lightweight wrapper around [sch_validate()] that
also returns the input.
}
\examples{
schema = sch_schema(
    .desc = "MCMC draws",
    .relationships = ~ chain * draw * parameter,
    chain = sch_integer("Chain number"),
    draw = sch_integer("Draw number", bounds = c(1, Inf), closed = c(TRUE, FALSE)),
    parameter = sch_character("Parameter name"),
    value = sch_numeric("Parameter value")
)
d_raw = data.frame(chain = 1L, draw = 1:4, parameter = "mu", value = rnorm(4))
d = new_sch_df(d_raw, schema, class="mcmc_draws")
validate_sch_df(d)
str(d)
}
